# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
Base Skeleton Graph.

Simple example usage:
>>> # --- Object
>>> import skimage.data as data
>>> import skimage.util as util
>>> object_map = util.invert(data.horse())
>>> # --- SKL Map
>>> from skl_graph.type.skl_map import SKLMapFromObjectMap
>>> skl_map = SKLMapFromObjectMap(object_map)
>>> # --- SKL Graph
>>> from skl_graph.type.graph import skl_graph_t
>>> skl_graph = skl_graph_t.NewFromSKLMap(skl_map)
>>> # --- Plotting
>>> import matplotlib.pyplot as pyplot
>>> _, all_axes = pyplot.subplots(ncols=3)
>>> all_axes[0].matshow(object_map, cmap="gray")
>>> all_axes[1].matshow(skl_map, cmap="gray")
>>> from skl_graph.task.plot.graph import Plot
>>> Plot(axes=all_axes[2], should_block=False)
>>> for axes, title in zip(all_axes, ("Object", "Skeleton", "Graph")):
>>>     axes.set_title(title)
>>>     axes.set_axis_off()
>>> pyplot.tight_layout()
>>> pyplot.show()
"""

from __future__ import annotations

import inspect as nspt
from typing import Any, Sequence

import networkx as ntkx
import numpy as nmpy
import scipy.ndimage as spim
import skimage.draw as skdw

import skl_graph.type.tpy_map as tgmp
from skl_graph.extension.string import AlignedNameAndValue
from skl_graph.type.edge import EdgesFromRawEdges, RawEdges, edge_t
from skl_graph.type.node import BranchNodes, EndNodes, branch_node_t, end_node_t
from skl_graph.type.style import (
    direction_style_raw_h,
    direction_style_t,
    edge_style_raw_h,
    edge_style_t,
    edge_styles_h,
    edge_styles_raw_h,
    label_style_raw_h,
    label_style_t,
    label_styles_h,
    label_styles_raw_h,
    node_either_raw_h,
    node_style_t,
    node_styles_h,
)

array_t = nmpy.ndarray


class skl_graph_t(ntkx.MultiGraph):
    """
    s_node: Singleton node
    e_node: End node
    b_node: Branch node
    """

    __slots__ = (
        "domain_shape",
        "n_components",
        "n_s_nodes",
        "n_e_nodes",
        "n_b_nodes",
        "has_widths",
        "node_styles",
        "edge_styles",
        "direction_style",
        "label_styles",
    )
    domain_shape: tuple[int, ...]
    n_components: int
    n_s_nodes: int
    n_e_nodes: int
    n_b_nodes: int
    has_widths: bool
    node_styles: node_styles_h  # One style per degree + one default style
    edge_styles: edge_styles_h  # One style for "regular" edges, one for self-loops
    direction_style: direction_style_t
    label_styles: label_styles_h  # For nodes and edges, in that order

    def __init__(self) -> None:
        """"""
        super().__init__()
        for slot in skl_graph_t.__slots__:
            setattr(self, slot, None)

    @classmethod
    def NewFromSKLMap(
        cls, skl_map: array_t, /, *, width_map: array_t = None
    ) -> skl_graph_t:
        """

        Parameters
        ----------
        skl_map : numpy.ndarray
        width_map : numpy.ndarray, optional

        Returns
        -------
        skl_graph_t
            Graph of the skeleton as an extended networkx.MultiGraph instance
        """
        output = cls()

        output.domain_shape = skl_map.shape
        output.has_widths = width_map is not None
        output.n_s_nodes = 0
        output.n_e_nodes = 0
        output.n_b_nodes = 0

        tmap, background_label = tgmp.TopologyMapOfMap(skl_map, return_bg_label=True)
        cc_map, n_components = tgmp.LABELING_FCT_FOR_DIM[skl_map.ndim](skl_map)

        output.n_components = n_components

        # Process skl_map/tmap per connected component (*)
        for cmp_label in range(1, n_components + 1):
            if n_components > 1:
                single_skl_map = cc_map == cmp_label
                single_tmap = nmpy.full_like(tmap, background_label)
                single_tmap[single_skl_map] = tmap[single_skl_map]
            else:
                single_skl_map = skl_map > 0
                single_tmap = tmap

            if output._DealsWithSpecialCases(
                single_tmap, background_label, width_map=width_map
            ):
                pass
            else:
                e_nodes, e_node_lmap = EndNodes(single_tmap, width_map=width_map)
                b_nodes, b_node_lmap = BranchNodes(single_tmap, width_map=width_map)
                raw_edges, edge_lmap = RawEdges(single_skl_map, b_node_lmap)
                edges, node_uids_per_edge = EdgesFromRawEdges(
                    raw_edges,
                    e_nodes,
                    b_nodes,
                    edge_lmap,
                    e_node_lmap,
                    b_node_lmap,
                    width_map=width_map,
                )

                output.add_nodes_from((_nde.uid, {"details": _nde}) for _nde in e_nodes)
                output.add_nodes_from((_nde.uid, {"details": _nde}) for _nde in b_nodes)
                for edge, adjacent_node_uids in zip(edges, node_uids_per_edge):
                    output.AddEdge(edge, adjacent_node_uids)

                output.n_e_nodes += e_nodes.__len__()
                output.n_b_nodes += b_nodes.__len__()

        output.node_styles = node_style_t.Default()
        output.edge_styles = edge_style_t.Default()
        output.direction_style = direction_style_t.Default()
        output.label_styles = label_style_t.Default()

        return output

    def _DealsWithSpecialCases(
        self, tmap: array_t, background_label: int, /, *, width_map: array_t = None
    ) -> bool:
        """Creates and adds nodes and edges of cases such as a singleton node, self loops...

        Parameters
        ----------
        tmap : numpy.ndarray
            Topological map of the skeleton; Must contain a unique connected component.
        background_label
        width_map

        Returns
        -------

        """
        singleton = nmpy.where(tmap == 0)
        if singleton[0].size > 0:
            # Can only be 1 since tmap is processed per connected components (*)
            singleton = nmpy.array(singleton, dtype=nmpy.int64).squeeze()
            _ = self.AddEndNode(singleton, width_map=width_map, is_singleton=True)

            return True
        #
        elif (tmap[tmap != background_label] == 2).all():
            # The tmap represents a self loop
            loop_slc = nmpy.nonzero(tmap == 2)

            # Takes the first pixel to serve as a node, and the rest for the self-loop edge
            #
            # 0:1 makes sites elements array_t's (instead of Numpy numbers), which is necessary for
            # branch_node_t.NewWithSites.
            sites = tuple(per_dim[0:1] for per_dim in loop_slc)
            node_uid = self.AddBranchNode(sites, width_map=width_map)

            n_unique_sites = loop_slc[0].__len__()
            sites = tuple(
                per_dim[list(range(n_unique_sites)) + [0]] for per_dim in loop_slc
            )
            adjacent_node_uids = (node_uid, node_uid)
            edge = edge_t.NewWithDetails(
                sites,
                False,
                adjacent_node_uids,
                width_map=width_map,
            )
            self.AddEdge(edge, adjacent_node_uids)

            return True

        return False

    def AddEndNode(
        self,
        position: array_t,
        /,
        *,
        width_map: array_t = None,
        is_singleton: bool = False,
    ) -> Any:
        """"""
        node = end_node_t.NewWithPosition(position, width_map=width_map)
        self.add_node(node.uid, details=node)

        if is_singleton:
            self.n_s_nodes += 1

        return node.uid

    def AddBranchNode(
        self,
        sites: tuple[array_t, ...],
        /,
        *,
        width_map: array_t = None,
    ) -> Any:
        """"""
        node = branch_node_t.NewWithSites(sites, width_map=width_map)
        self.add_node(node.uid, details=node)

        self.n_b_nodes += 1

        return node.uid

    def AddEdge(self, edge: edge_t, adjacent_node_uids: Sequence[str], /) -> None:
        """"""
        edge_uid = edge.uid
        version_number = 0
        uid_w_vn = edge_uid
        while self.has_edge(*adjacent_node_uids, key=uid_w_vn):
            version_number += 1
            uid_w_vn = edge_uid + "+" + version_number.__str__()

        # The key is added to distinguish between multiple "parallel" edges. As a consequence, an edge is accessed as
        # follows: self.edges[adjacent_node_uids[0], adjacent_node_uids[1], uid_w_vn]
        self.add_edge(*adjacent_node_uids, key=uid_w_vn, details=edge)

    @property
    def dim(self) -> int:
        """"""
        return self.domain_shape.__len__()

    @property
    def n_nodes(self) -> int:
        """"""
        return self.number_of_nodes()

    @property
    def n_edges(self) -> int:
        """"""
        return self.number_of_edges()

    @property
    def n_self_loops(self) -> int:
        """"""
        return ntkx.number_of_selfloops(self)

    @property
    def degree_statistics(self) -> tuple[int, int, float, float, float, array_t]:
        """"""
        degrees = tuple(degree for _, degree in self.degree)
        return (
            min(degrees),
            max(degrees),
            nmpy.median(degrees).item(),
            nmpy.mean(degrees).item(),
            nmpy.std(degrees).item(),
            nmpy.bincount(degrees),
        )

    @property
    def least_and_most_connected_nodes(
        self,
    ) -> tuple[tuple[str, ...], int, tuple[str, ...], int]:
        """"""
        min_degree, max_degree, *_ = self.degree_statistics
        least = (_nde for _nde, _dgr in self.degree if _dgr == min_degree)
        most = (_nde for _nde, _dgr in self.degree if _dgr == max_degree)

        return tuple(least), min_degree, tuple(most), max_degree

    @property
    def length(self) -> float:
        """"""
        return sum(_lgt for *_, _lgt in self.lengths)

    @property
    def lengths(self) -> tuple[tuple[str, str, str, float], ...]:
        """"""
        # TODO: check that the newly retrieved uid does not break calls.
        return tuple(
            (_src, _tgt, _uid, _dtl.lengths.length)
            for _src, _tgt, _uid, _dtl in self.edges.data("details", keys=True)
        )

    @property
    def mean_based_area(self) -> float:
        """
        See self.mean_based_areas
        """
        return sum(_are for *_, _are in self.mean_based_areas)

    @property
    def mean_based_areas(self) -> tuple[tuple[str, str, str, float], ...]:
        """
        mean=using mean edge widths
        """
        # TODO: check that the newly retrieved uid does not break code.
        if self.has_widths:
            return tuple(
                (_src, _tgt, _uid, _dtl.areas.mean_based_area)
                for _src, _tgt, _uid, _dtl in self.edges.data("details", keys=True)
            )

        return self.n_edges * (("", "", "", 0.0),)

    @property
    def segment_based_area(self) -> float:
        """
        See self.segment_based_areas
        """
        return sum(_are for *_, _are in self.segment_based_areas)

    @property
    def segment_based_areas(self) -> tuple[tuple[str, str, str, float], ...]:
        """
        segment=using edge segments
        """
        # TODO: check that the newly retrieved uid does not break code.
        if self.has_widths:
            return tuple(
                (_src, _tgt, _uid, _dtl.areas.segment_based_area)
                for _src, _tgt, _uid, _dtl in self.edges.data("details", keys=True)
            )

        return self.n_edges * (("", "", "", 0.0),)

    @property
    def bbox(self) -> tuple[int, ...]:
        """"""
        bbox_min = nmpy.full(self.dim, max(self.domain_shape), dtype=nmpy.int64)
        bbox_max = nmpy.zeros(self.dim, dtype=nmpy.int64)
        for *_, edge in self.edges.data("details"):
            sites = nmpy.array(edge.sites, dtype=nmpy.int64)
            bbox_min = nmpy.fmin(bbox_min, nmpy.amin(sites, axis=1))
            bbox_max = nmpy.fmax(bbox_max, nmpy.amax(sites, axis=1))

        output = []
        for d_idx in range(self.dim):
            output.extend((bbox_min[d_idx], bbox_max[d_idx]))

        return tuple(output)

    def AsNetworkX(self) -> ntkx.MultiGraph:
        """"""
        return ntkx.MultiGraph(incoming_graph_data=self)
        # TODO: Why storing uid in edge while it is already available as source, target, uid?
        # output = ntkx.MultiGraph(incoming_graph_data=self)
        #
        # for source, target, details in self.edges.data("details"):
        #     uid = details.uid
        #     output.edges[source, target, uid]["uid"] = uid
        #
        # return output

    def RebuiltSkeletonMap(self, /, *, with_width: bool = False) -> array_t:
        """"""
        if (not self.has_widths) and with_width:
            with_width = False
        if with_width:
            dtype = nmpy.float64
        else:
            # Not uint to allow for subtraction
            dtype = nmpy.int8

        output = nmpy.zeros(self.domain_shape, dtype=dtype)

        for *_, details in self.edges.data("details"):
            if with_width:
                output[details.sites] = details.widths
            else:
                output[details.sites] = 2

        for _, details in self.nodes.data("details"):
            if isinstance(details, branch_node_t):
                if with_width:
                    output[details.sites] = details.diameters
                else:
                    output[details.sites] = 3
            else:
                if with_width:
                    output[tuple(details.position)] = details.diameter
                else:
                    output[tuple(details.position)] = 1

        return output

    def RebuiltObjectMap(self) -> array_t:
        """"""
        if not self.has_widths:
            raise ValueError("Requires an SKL graph with widths")

        # Not uint to allow for subtraction
        output = nmpy.zeros(self.domain_shape, dtype=nmpy.int8)

        if self.dim == 2:
            NewBall = skdw.disk
        else:
            NewBall = _Ball3D

        for _, details in self.nodes.data("details"):
            if isinstance(details, branch_node_t):
                for *sites, radius in zip(
                    *details.sites,
                    nmpy.around(0.5 * (details.diameters - 1.0)).astype(nmpy.int64),
                ):
                    output[NewBall(sites, radius, shape=output.shape)] = 1
            else:
                output[
                    NewBall(
                        details.position,
                        nmpy.around(0.5 * (details.diameter - 1.0))
                        .astype(nmpy.int64)
                        .item(),
                        shape=output.shape,
                    )
                ] = 1

        for *_, details in self.edges.data("details"):
            for *sites, radius in zip(
                *details.sites,
                nmpy.around(0.5 * (details.widths - 1.0)).astype(nmpy.int64),
            ):
                output[NewBall(sites, radius, shape=output.shape)] = 1

        return output

    def SetPlotStyles(
        self,
        /,
        *,
        node: node_either_raw_h = None,
        #
        edge: edge_style_raw_h = None,
        regular_edge: edge_style_raw_h = None,
        self_loop: edge_style_raw_h = None,
        edges: edge_styles_raw_h = None,
        #
        label: label_style_raw_h = None,
        label_node: label_style_raw_h = None,
        label_edge: label_style_raw_h = None,
        labels: label_styles_raw_h = None,
        #
        direction: direction_style_raw_h = None,
    ) -> None:
        """
        edge: common style for both regulars and self-loops
        edges: styles for regulars and self-loops, in that order
        label: common style for both nodes and edges
        labels: styles for nodes and edges, in that order
        """
        if node is not None:
            styles = node_style_t.NewFromUnstructured(node)
            if isinstance(node, dict):  # node_styles_raw_h
                self.node_styles.update(styles)
            else:
                self.node_styles = styles
            if None not in self.node_styles:
                # Should never happen, but just in case...
                raise ValueError(
                    f"{self.node_styles}: Node styles without default style. "
                    f'Expected: style dictionary with a "None" entry.'
                )

        styles = edge_style_t.AllFromUnstructured(
            self.edge_styles,
            edge=edge,
            regular_edge=regular_edge,
            self_loop=self_loop,
            edges=edges,
        )
        if styles is not None:
            self.edge_styles = styles

        styles = label_style_t.AllFromUnstructured(
            self.label_styles,
            label=label,
            node=label_node,
            edge=label_edge,
            labels=labels,
        )
        if styles is not None:
            self.label_styles = styles

        if direction is not None:
            self.direction_style = direction_style_t.NewFromUnstructured(direction)

    def Properties(self, /, *, prefix: str = "") -> str:
        """"""
        output = []

        max_name_length = 0
        # The loop below seems unnecessarily complicated. But the simpler call
        # "nspt.getmembers(self, lambda _elm: isinstance(_elm, property))" does not work since _elm is the actual value,
        # so...
        for name, value in nspt.getmembers(self):
            # Somehow, a name returned by getmembers might not be an attribute of the class !
            attribute = getattr(self.__class__, name, None)
            if attribute is None:
                continue

            if isinstance(attribute, property) and not isinstance(value, Sequence):
                if (current_length := name.__len__()) > max_name_length:
                    max_name_length = current_length
                output.append(f"{name}: {value}")
        output = map(
            lambda _elm: AlignedNameAndValue(_elm, max_name_length + 1, prefix), output
        )

        return "\n".join(output)

    def __str__(self) -> str:
        """"""
        output = (
            f"{self.__class__.__name__}:\n"
            f"    Domain shape: {self.domain_shape}\n"
            f"    Has widths:   {self.has_widths}\n\n"
            f"    Components:   {self.n_components}\n"
            f"    Nodes:        {self.n_nodes}"
            f" = S_{self.n_s_nodes} + E_{self.n_e_nodes} + B_{self.n_b_nodes}\n"
            f"    Edges:        {self.n_edges}, among which {self.n_self_loops} self loop(s)"
        )

        return output


def _Ball3D(
    center: Sequence[int], radius: int, /, *, shape: tuple[int, int, int] = None
) -> array_t:
    """
    Signature follows the one of skimage.draw.disk
    """
    output = nmpy.zeros(shape, dtype=nmpy.bool_)
    # skdw.ellipsoid leaves a one pixel margin around the ellipse, hence [1:-1, 1:-1, 1:-1]
    ellipse = skdw.ellipsoid(radius, radius, radius)[1:-1, 1:-1, 1:-1]
    sp_slices = tuple(
        slice(0, min(output.shape[idx_], ellipse.shape[idx_])) for idx_ in (0, 1, 2)
    )
    output[sp_slices] = ellipse[sp_slices]

    row, col, dep = center
    output = spim.shift(
        output, (row - radius, col - radius, dep - radius), order=0, prefilter=False
    )

    return output



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Sequence

import networkx as ntkx
import numpy as nmpy

import skl_graph.type.tpy_map as tgmp
from skl_graph.task.check.edge import Issues as EdgeIssues
from skl_graph.task.check.node import Issues as NodeIssues
from skl_graph.type.graph import skl_graph_t

array_t = nmpy.ndarray


def Issues(graph: skl_graph_t, /) -> Sequence[str]:
    """"""
    # TODO: Check all attributes values, like with nodes, and do it first.
    output = []

    n_components = ntkx.number_connected_components(graph)
    if n_components != graph.n_components:
        output.append(
            f"Actual and stored number of connected components differ: "
            f"{n_components}!={graph.n_components}"
        )
    n_s_nodes = sum(degree == 0 for _, degree in graph.degree)
    if n_s_nodes != graph.n_s_nodes:
        output.append(
            f"Actual and stored number of singleton nodes differ: "
            f"{n_s_nodes}!={graph.n_s_nodes}"
        )
    n_e_nodes = sum(degree == 1 for _, degree in graph.degree)
    if n_e_nodes != graph.n_e_nodes:
        output.append(
            f"Actual and stored number of end nodes differ: "
            f"{n_e_nodes}!={graph.n_e_nodes}"
        )
    n_b_nodes = sum(degree > 1 for _, degree in graph.degree)
    if n_b_nodes != graph.n_b_nodes:
        output.append(
            f"Actual and stored number of end nodes differ: "
            f"{n_b_nodes}!={graph.n_b_nodes}"
        )

    for _, details in graph.nodes.data("details"):
        output.extend(NodeIssues(details))

    for source, target, uid, details in graph.edges.data("details", keys=True):
        source, target = (graph.nodes[_uid]["details"] for _uid in (source, target))
        output.extend(EdgeIssues(details, source, target, uid))

    return output


def Correctness(graph: skl_graph_t, skl_map: array_t, /) -> tuple[bool, bool]:
    """"""
    rebuilt_skl_map = graph.RebuiltSkeletonMap()

    topological_map = tgmp.TopologyMapOfMap(skl_map)
    # Keep next line before its next one
    topological_map[topological_map == tgmp.TMapBackgroundLabel(skl_map)] = 0
    topological_map[topological_map > 3] = 3

    binary_correctness = nmpy.array_equal(rebuilt_skl_map > 0, skl_map)
    topology_correctness = nmpy.array_equal(rebuilt_skl_map, topological_map)

    return binary_correctness, topology_correctness



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

# skl_ograph=Skeleton graph with specific operations, derived from skeleton graph with features

import networkx as ntkx
import numpy as nmpy

from skl_graph.type.edge import edge_t
from skl_graph.type.graph import skl_graph_t
from skl_graph.type.node import end_node_t

print(f"--- {__name__}: THIS MODULE IS A WORK-IN-PROGRESS: NOT USABLE YET ---")


# Derive an oedge_t (or other name) adding a field composition:
# composition: Holds the lengths of pieces composing the edge (can change while simplifying)


array_t = nmpy.ndarray


def PruneBasedOnWidths(graph: skl_graph_t, min_width: float, /) -> None:
    """"""
    if not graph.has_widths:
        raise ValueError("Requires an SKL graph with widths")

    delete_list = []
    relabeling_dct = {}

    for source, target, uid, details in graph.edges.data("details", keys=True):
        # TODO: uid now retrieved but unhandled.
        extremity = None
        if graph.degree[source] == 1:
            extremity = source
        elif graph.degree[target] == 1:
            extremity = target
        if extremity is None:
            continue

        if details["composition"].__len__() != 1:
            raise ValueError("Not made to work on already simplified graphs")

        edge_coords_0 = details["sites"][0]
        edge_coords_1 = details["sites"][1]
        widths = details["widths"]
        n_edge_pixels = len(edge_coords_0)

        if (edge_coords_0[0], edge_coords_1[0]) == graph.nodes[extremity].position:
            pixel_idx = 0
            idx_incr = 1
        else:
            pixel_idx = n_edge_pixels - 1
            idx_incr = -1

        while (0 <= pixel_idx < n_edge_pixels) and (widths[pixel_idx] < min_width):
            pixel_idx += idx_incr

        if (pixel_idx < 0) or (pixel_idx >= n_edge_pixels):
            delete_list.append(extremity)
        else:
            graph.nodes[extremity].position = (
                edge_coords_0[pixel_idx],
                edge_coords_1[pixel_idx],
            )
            graph.nodes[extremity].diameter = widths[pixel_idx]
            unused = end_node_t.NewWithPosition(graph.nodes[extremity].position)
            relabeling_dct[extremity] = unused.uid

            if idx_incr > 0:
                valid_idc = slice(pixel_idx, n_edge_pixels)
                extra_idc = slice(pixel_idx + 1)
            else:
                valid_idc = slice(pixel_idx + 1)
                extra_idc = slice(pixel_idx, n_edge_pixels)

            edge_piece = nmpy.array(
                (edge_coords_0[extra_idc], edge_coords_1[extra_idc]), dtype=nmpy.float64
            )
            extra_widths = nmpy.array(widths[extra_idc], dtype=nmpy.float64)
            extra_lengths = nmpy.sqrt((nmpy.diff(edge_piece, axis=1) ** 2).sum(axis=0))
            extra_w_lengths = extra_lengths * (
                0.5 * (extra_widths[1:] + extra_widths[:-1])
            )

            details["sites"] = (
                edge_coords_0[valid_idc],
                edge_coords_1[valid_idc],
            )
            details["widths"] = widths[valid_idc]
            details[
                "length"
            ] -= extra_lengths.sum().item()  # Conversion to float is necessary,
            details[
                "w_length"
            ] -= (
                extra_w_lengths.sum().item()
            )  # otherwise type nmpy.float64 contaminates all.
            if details["origin_node"] == extremity:
                details["origin_node"] = relabeling_dct[extremity]

    if len(delete_list):
        graph.remove_nodes_from(delete_list)
    if len(relabeling_dct):
        ntkx.relabel_nodes(graph, relabeling_dct, copy=False)


def Simplify(graph: skl_graph_t, min_edge_length: float, /) -> None:
    """"""
    while True:
        min_length = nmpy.Inf
        end_nodes = ()
        for source, target, uid, edge_details in graph.edges.data("details", keys=True):
            # TODO: uid now retrieved but unhandled.
            degree_0 = graph.degree[source]
            degree_1 = graph.degree[target]

            if (degree_0 > 2) and (degree_1 > 2) and (edge_details.length < min_length):
                edge_desc_list = graph[source][target]
                if len(edge_desc_list) > 1:
                    lengths_lower = [
                        description["length"] <= edge_details.length
                        for description in edge_desc_list.values()
                    ]
                    should_continue = all(lengths_lower)
                else:
                    should_continue = True
                if should_continue:
                    min_length = edge_details.length
                    end_nodes = (source, target)

        if min_length < min_edge_length:
            # /!\ management of edge descriptions is inexistant
            edges = []
            all_coords = [[], []]
            n_coords_per_piece = []
            for edge_uid, edge_details in graph[end_nodes[0]][end_nodes[1]].items():
                edges.append(edge_uid)
                sites = edge_details["sites"]
                all_coords[0].extend(sites[0])
                all_coords[1].extend(sites[1])
                n_coords_per_piece.append(len(sites[0]))

            all_coords = nmpy.array(all_coords, dtype=nmpy.float64)
            # cum_n_coords_per_piece = nmpy.cumsum(n_coords_per_piece) - 1
            # Naming: actually "cumulative minus one" rather than "cumulative"

            centroid = all_coords.mean(axis=1, keepdims=True).round()
            closest_pixel_idx = (
                ((all_coords - centroid) ** 2).sum(axis=0).argmin()
            )  # idx of first occurrence of min
            centroid = (
                int(all_coords[0, closest_pixel_idx]),
                int(all_coords[1, closest_pixel_idx]),
            )

            # closest_edge_idx = cum_n_coords_per_piece.searchsorted(
            #     closest_pixel_idx
            # )
            # shared_description = graph[end_nodes[0]][end_nodes[1]][
            #     edges[closest_edge_idx]
            # ]
            # for edge_uid in graph[end_nodes[0]][end_nodes[1]].keys():
            #     if edge_uid != edge_to_keep:
            #         graph.remove_edge(end_nodes[0], end_nodes[1], key = edge_uid)

            description = graph.nodes[end_nodes[0]]
            description["position"] = centroid
            description["sites"] = ((centroid[0],), (centroid[1],))

            # for neighbor in graph[end_nodes[0]]:
            #     if neighbor != end_nodes[1]:
            #         for description in graph[end_nodes[0]][neighbor].values():
            #             pass
            #             # use shared_description here
            #             # description = {
            #             #     "sites":           new_coords,
            #             #     'origin_node': edge_0_node,
            #             #     'length':           edge_0_desc['length']   + edge_1_desc['length']   + joint_length,
            #             #     'w_length':         edge_0_desc['w_length'] + edge_1_desc['w_length'] + joint_w_length,
            #             #     'widths':            widths,
            #             #     'composition':      edge_0_desc['composition'] + (len(node_details["sites"][0]),) + \
            #             #                         edge_1_desc['composition']
            #             # }

            for neighbor in graph[end_nodes[1]]:
                if (neighbor != end_nodes[0]) and (neighbor not in graph[end_nodes[0]]):
                    # for edge_uid, edge_details in graph[end_nodes[1]][neighbor].items():
                    # use shared_description here
                    # description = {
                    #     "sites":           new_coords,
                    #     'origin_node': edge_0_node,
                    #     'length':           edge_0_desc['length']   + edge_1_desc['length']   + joint_length,
                    #     'w_length':         edge_0_desc['w_length'] + edge_1_desc['w_length'] + joint_w_length,
                    #     'widths':            widths,
                    #     'composition':      edge_0_desc['composition'] + (len(node_details["sites"][0]),) + \
                    #                         edge_1_desc['composition']
                    # }
                    # graph.add_edge(
                    #     end_nodes[0], neighbor, key=edge_uid, **edge_details
                    # )
                    edge = edge_t.NewWithDetails(
                        (nmpy.array([]), nmpy.array([])),
                        False,
                        (end_nodes[0], neighbor),
                    )
                    graph.AddEdge(edge, (end_nodes[0], neighbor))

            graph.remove_node(end_nodes[1])
        else:
            break

    while True:
        min_length = nmpy.Inf
        node_uid = -1
        for source, target, uid, edge_details in graph.edges.data("details", keys=True):
            # TODO: uid now retrieved but unused.
            degree_0 = graph.degree[source]
            degree_1 = graph.degree[target]

            if (
                ((degree_0 == 1) or (degree_1 == 1))
                and (degree_0 + degree_1 > 3)
                and (edge_details.length < min_length)
            ):
                min_length = edge_details.length
                if degree_0 == 1:
                    node_uid = source
                else:
                    node_uid = target

        if min_length < min_edge_length:
            graph.remove_node(node_uid)
        else:
            break

    graph_has_been_modified = True
    while graph_has_been_modified:
        graph_has_been_modified = False

        for node_uid, node_details in graph.nodes.data("details"):
            if graph.degree[node_uid] != 2:
                continue

            edge_0, edge_1 = graph.edges(node_uid, data=True)
            other_node_0, edge_0_node, edge_0_desc = edge_0
            other_node_1, edge_1_node, edge_1_desc = edge_1
            assert (other_node_0 == node_uid) and (other_node_1 == node_uid)
            # If this assertion fails one day, it means that NetworkX has changed the way it returns
            # adjacent edges. It will become necessary to test which of other_node_X and edge_X_node
            # is node_uid.

            new_coords, joint_length, first_reversed, last_reversed = _EdgeOfGluedEdges(
                edge_0_desc["sites"],
                edge_1_desc["sites"],
                node_details["sites"],
                edge_0_desc["origin_node"] == node_uid,
                edge_1_desc["origin_node"] == node_uid,
            )

            # if graph.has_widths:
            #     joint_w_length = joint_length * nmpy.mean(node_details["diameters"])
            #     if first_reversed:
            #         first_part = tuple(reversed(edge_0_desc["widths"]))
            #     else:
            #         first_part = edge_0_desc["widths"]
            #     if last_reversed:
            #         last_part = tuple(reversed(edge_1_desc["widths"]))
            #     else:
            #         last_part = edge_1_desc["widths"]
            #     widths = first_part + node_details["diameters"] + last_part
            # else:
            #     joint_w_length = 0
            #     widths = None

            # description = {
            #     "sites": new_coords,
            #     "origin_node": edge_0_node,
            #     "length": edge_0_desc["length"]
            #     + edge_1_desc["length"]
            #     + joint_length,
            #     "w_length": edge_0_desc["w_length"]
            #     + edge_1_desc["w_length"]
            #     + joint_w_length,
            #     "widths": widths,
            #     "composition": edge_0_desc["composition"]
            #     + (len(node_details["sites"][0]),)
            #     + edge_1_desc["composition"],
            # }

            edge = edge_t.NewWithDetails(new_coords, True, (edge_0_node, edge_1_node))
            graph.AddEdge(edge, (edge_0_node, edge_1_node))
            # graph.add_edge(
            #     edge_0_node,
            #     edge_1_node,
            #     EdgeID(edge_0_node, edge_1_node),
            #     **description
            # )
            graph.remove_node(node_uid)
            graph_has_been_modified = True
            break


def _EdgeOfGluedEdges(
    edge_0_coords, edge_1_coords, node_coords, node_is_first_of_0, node_is_first_of_1, /
):
    """
    Returns the glued sites and the length of the gluing joint
    """
    if node_is_first_of_0:
        first_part_0 = tuple(reversed(edge_0_coords[0]))
        first_part_1 = tuple(reversed(edge_0_coords[1]))
        first_reversed = True
    else:
        first_part_0 = edge_0_coords[0]
        first_part_1 = edge_0_coords[1]
        first_reversed = False

    if node_is_first_of_1:
        last_part_0 = tuple(reversed(edge_1_coords[0]))
        last_part_1 = tuple(reversed(edge_1_coords[1]))
        last_reversed = True
    else:
        last_part_0 = edge_1_coords[0]
        last_part_1 = edge_1_coords[1]
        last_reversed = False

    glued_coords = (
        first_part_0 + node_coords[0] + last_part_0,
        first_part_1 + node_coords[1] + last_part_1,
    )
    joint_length = float(
        nmpy.sqrt(
            (first_part_0[-1] - last_part_0[0]) ** 2
            + (first_part_1[-1] - last_part_1[0]) ** 2
        )
    )

    return glued_coords, joint_length, first_reversed, last_reversed



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Sequence

import numpy as nmpy

from skl_graph.extension.numpy_ import Issues as NumpyArrayIssues
from skl_graph.type.node import branch_node_t, end_node_t

array_t = nmpy.ndarray


def Issues(node: end_node_t | branch_node_t, /) -> Sequence[str]:
    """"""
    output = []

    issues = NumpyArrayIssues(
        node.position,
        expected_dtype=nmpy.integer,
        expected_size=(2, 3),
        expected_shape=((2,), (3,)),
    )
    if issues.__len__() > 0:
        issues = ("Node position: " + _iss for _iss in issues)
        output.extend(issues)
        expected_dim = None
    else:
        expected_dim = node.position.size

    if node.uid is None:
        output.append("Node without UID")

    if isinstance(node, branch_node_t):
        if isinstance(node.sites, tuple) and (node.sites.__len__() in (2, 3)):
            if (expected_dim is None) or (node.sites.__len__() == expected_dim):
                if all(isinstance(_elm, array_t) for _elm in node.sites):
                    expected_size = node.sites[0].size
                    expected_shape = (expected_size,)
                    for c_idx, component in enumerate(node.sites):
                        issues = NumpyArrayIssues(
                            component,
                            expected_dtype=nmpy.integer,
                            expected_size=expected_size,
                            expected_shape=expected_shape,
                        )
                        if issues.__len__() > 0:
                            issues = (f"Node sites[{c_idx}]: {_iss}" for _iss in issues)
                            output.extend(issues)
                else:
                    output.append(
                        f"{tuple(type(_elm).__name__ for _elm in node.sites)}: "
                        f"Invalid type(s) of node sites components. "
                        f"Expected=all {array_t.__name__}'s."
                    )
            else:
                output.append(
                    f"{node.sites.__len__()}: Invalid node sites dimension. "
                    f"Expected={expected_dim}."
                )
        else:
            if isinstance(node.sites, Sequence):
                dimension = node.sites.__len__()
            else:
                dimension = "N/A"
            output.append(
                f"T:{type(node.sites).__name__}, D:{dimension}: "
                f"Invalid type T or dimension D of node sites. "
                f"Expected=tuple with 2 or 3 elements."
            )

    return output



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Iterable

import networkx as ntkx
import numpy as nmpy
from matplotlib.collections import PatchCollection as patch_2d_collection_t
from matplotlib.patches import Rectangle as rectangle_t
from mpl_toolkits.mplot3d.art3d import Poly3DCollection as poly_3d_collection_t

from skl_graph.task.plot.base import axes_3d_t, axes_t, y_transform_h
from skl_graph.type.graph import skl_graph_t
from skl_graph.type.node import array_t, branch_node_t, end_node_t, node_t
from skl_graph.type.style import label_style_t, node_styles_h

# Note: these array_t's are one-element arrays, so the type should be tuple[int, ...] if it were not for not wasting
# time with .item()'s in the elements of TransformedPosition.
positions_as_dict_h = dict[str, tuple[array_t, ...]]


_CUBE = [
    [[-0.5, 0.5, -0.5], [-0.5, -0.5, -0.5], [0.5, -0.5, -0.5], [0.5, 0.5, -0.5]],
    [[-0.5, -0.5, -0.5], [-0.5, -0.5, 0.5], [0.5, -0.5, 0.5], [0.5, -0.5, -0.5]],
    [[0.5, -0.5, 0.5], [0.5, -0.5, -0.5], [0.5, 0.5, -0.5], [0.5, 0.5, 0.5]],
    [[-0.5, -0.5, 0.5], [-0.5, -0.5, -0.5], [-0.5, 0.5, -0.5], [-0.5, 0.5, 0.5]],
    [[-0.5, 0.5, -0.5], [-0.5, 0.5, 0.5], [0.5, 0.5, 0.5], [0.5, 0.5, -0.5]],
    [[-0.5, 0.5, 0.5], [-0.5, -0.5, 0.5], [0.5, -0.5, 0.5], [0.5, 0.5, 0.5]],
]
_CUBE = nmpy.array(_CUBE)
_CUBE_TRANSPARENCY = 0.6


def PositionsForPlotFromDetails(
    details: Iterable[tuple[str, node_t]], TransformedY: y_transform_h, /
) -> positions_as_dict_h:
    """"""
    TransformedPosition = lambda _psn: (_psn[1], TransformedY(_psn[0]), *_psn[2:])

    return dict((_uid, TransformedPosition(_dtl.position)) for _uid, _dtl in details)


def Plot2DEndNodes(
    nodes: Iterable[tuple[str, node_t]],
    TransformedY: y_transform_h,
    axes: axes_t,
    node_styles: node_styles_h,
    /,
) -> None:
    """
    nodes: all the nodes of the graph, so that filtering is needed here
    """
    style_set = set(node_styles.keys())
    squares_and_styles = [
        (
            rectangle_t(
                (_dtl.position[1] - 0.5, TransformedY(_dtl.position[0]) - 0.5), 1, 1
            ),
            style_set.intersection(_dtl.properties.keys()),
        )
        for _, _dtl in nodes
        if isinstance(_dtl, end_node_t)
    ]

    if squares_and_styles.__len__() > 0:
        squares, custom_styles = zip(*squares_and_styles)
        degree_based_color = node_styles.get(1, node_styles[None]).color
        node_colors = tuple(
            _sty.pop().color if _sty.__len__() > 0 else degree_based_color
            for _sty in custom_styles
        )
        if set(node_colors).__len__() > 1:
            for square, color in zip(squares, node_colors):
                square.set_facecolor(color)
                axes.add_patch(square)
        else:
            collection = patch_2d_collection_t(squares, facecolor=node_colors[0])
            axes.add_collection(collection)


def Plot3DEndNodes(
    nodes: Iterable[tuple[str, node_t]],
    TransformedY: y_transform_h,
    axes: axes_3d_t,
    node_styles: node_styles_h,
    /,
) -> None:
    """
    nodes: all the nodes of the graph, so that filtering is needed here
    """
    style_set = set(node_styles.keys())
    cubes_and_styles = [
        (
            nmpy.add(
                _CUBE,
                (_dtl.position[1], TransformedY(_dtl.position[0]), _dtl.position[2]),
            ),
            style_set.intersection(_dtl.properties.keys()),
        )
        for _, _dtl in nodes
        if isinstance(_dtl, end_node_t)
    ]

    if cubes_and_styles.__len__() > 0:
        cubes, custom_styles = zip(*cubes_and_styles)
        degree_based_color = node_styles.get(1, node_styles[None]).color
        node_colors = tuple(
            _sty.pop().color if _sty.__len__() > 0 else degree_based_color
            for _sty in custom_styles
        )
        if set(node_colors).__len__() > 1:
            for cube, color in zip(cubes, node_colors):
                cube.set_facecolor(color)
                cube.set_alpha(_CUBE_TRANSPARENCY)
                axes.add_patch(cube)
        else:
            collection = poly_3d_collection_t(
                nmpy.concatenate(cubes),
                facecolor=node_colors[0],
                alpha=_CUBE_TRANSPARENCY,
            )
            axes.add_collection3d(collection)


def Plot2DBranchNodes(
    nodes: Iterable[tuple[str, node_t]],
    degrees: dict[str, int],
    TransformedY: y_transform_h,
    axes: axes_t,
    node_styles: node_styles_h,
    /,
) -> None:
    """
    nodes: all the nodes of the graph, so that filtering is needed here
    """
    default_style = node_styles[None]
    style_set = set(node_styles.keys())
    for uid, details in nodes:
        if isinstance(details, branch_node_t):
            custom_style = style_set.intersection(details.properties.keys())
            if custom_style.__len__() > 0:
                node_style = custom_style.pop()
            else:
                node_style = degrees[uid]
            node_style = node_styles.get(node_style, default_style)

            corners_0 = details.sites[1] - 0.5
            corners_1 = TransformedY(details.sites[0]) - 0.5
            squares = [
                rectangle_t(corner, 1, 1) for corner in zip(corners_0, corners_1)
            ]

            collection = patch_2d_collection_t(squares, facecolor=node_style.color)
            axes.add_collection(collection)


def Plot3DBranchNodes(
    nodes: Iterable[tuple[str, node_t]],
    degrees: dict[str, int],
    TransformedY: y_transform_h,
    axes: axes_3d_t,
    node_styles: node_styles_h,
    /,
) -> None:
    """
    nodes: all the nodes of the graph, so that filtering is needed here
    """
    default_style = node_styles[None]
    style_set = set(node_styles.keys())
    for uid, details in nodes:
        if isinstance(details, branch_node_t):
            custom_style = style_set.intersection(details.properties.keys())
            if custom_style.__len__() > 0:
                node_style = custom_style.pop()
            else:
                node_style = degrees[uid]
            node_style = node_styles.get(node_style, default_style)

            corners_0 = details.sites[1] - 0.5
            corners_1 = TransformedY(details.sites[0]) - 0.5
            corners_2 = details.sites[2] - 0.5
            cubes = [
                nmpy.add(_CUBE, corner)
                for corner in zip(corners_0, corners_1, corners_2)
            ]

            collection = poly_3d_collection_t(
                nmpy.concatenate(cubes),
                facecolor=node_style.color,
                alpha=_CUBE_TRANSPARENCY,
            )
            axes.add_collection3d(collection)


def Plot2DNodeLabels(
    skl_graph: skl_graph_t,
    node_label_positions: positions_as_dict_h,
    axes: axes_t,
    /,
) -> None:
    """"""
    # NetworkX changes axes styles. Save here to set back afterward.
    axes_styles = {}
    for axis, name in ((axes.xaxis, "x"), (axes.yaxis, "y")):
        axes_styles[name] = {}
        for level in ("minor", "major"):
            axes_styles[name][level] = axis.get_tick_params(which=level)

    ntkx.draw_networkx_labels(
        skl_graph,
        ax=axes,
        pos=node_label_positions,
        font_size=int(round(skl_graph.label_styles[0].size)),
        font_color=skl_graph.label_styles[0].color,
    )

    for name, styles in axes_styles.items():
        for level, style in styles.items():
            axes.tick_params(
                axis=name,
                which=level,
                **style,
            )


def Plot3DNodeLabels(
    nodes: Iterable[str],
    positions_as_dict: positions_as_dict_h,
    axes: axes_3d_t,
    style: label_style_t,
    /,
) -> None:
    """"""
    for node in nodes:
        axes.text(
            *positions_as_dict[node], node, fontsize=style.size, color=style.color
        )



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import Any

import numpy as nmpy
import skimage.measure as sims

import skl_graph.type.tpy_map as tgmp
from skl_graph.extension.number import COORDINATE_SEPARATOR

array_t = nmpy.ndarray


@dtcl.dataclass(slots=True, repr=False, eq=False)
class node_t:
    position: array_t
    uid: str = None
    properties: dict[Any, Any] = dtcl.field(init=False, default_factory=dict)

    def __post_init__(self) -> None:
        """"""
        self.uid = COORDINATE_SEPARATOR.join(map(str, self.position))

    def SetProperty(self, name: str, /, *, value: Any = None):
        """"""
        self.properties[name] = value

    @property
    def dim(self) -> int:
        """"""
        return self.position.__len__()

    def __str__(self) -> str:
        """"""
        return f"{self.__class__.__name__}.{self.uid} @ {self.position}"


@dtcl.dataclass(slots=True, repr=False, eq=False)
class end_node_t(node_t):
    diameter: float = None

    @classmethod
    def NewWithPosition(
        cls, position: array_t, /, *, width_map: array_t = None
    ) -> end_node_t:
        """"""
        output = cls(position=position)

        if width_map is not None:
            output.diameter = width_map.item(tuple(position))

        return output

    def __str__(self) -> str:
        """"""
        # super().__str__() produces:
        #     TypeError: super(type, obj): obj must be an instance or subtype of type
        # This is due to a combination of dataclass and slots=True.
        output = node_t.__str__(self)
        output += f"\n    Diameter: {self.diameter}"
        return output


@dtcl.dataclass(slots=True, repr=False, eq=False)
class branch_node_t(node_t):
    sites: tuple[array_t, ...] = None
    diameters: array_t = None

    @classmethod
    def NewWithSites(
        cls,
        sites: tuple[array_t, ...],
        /,
        *,
        width_map: array_t = None,
    ) -> branch_node_t:
        """"""
        sites_as_array = nmpy.array(sites).reshape((sites.__len__(), sites[0].size))
        centroid = nmpy.mean(sites_as_array, axis=1, keepdims=True)
        segments = sites_as_array - centroid
        medoid_idx = (segments**2).sum(axis=0).argmin()
        # nmpy.array(): fresh array instead of a view of sites_as_array
        position = nmpy.array(sites_as_array[:, medoid_idx].squeeze())

        output = cls(position=position)

        output.sites = sites
        if width_map is not None:
            output.diameters = width_map[sites]

        return output

    def __str__(self) -> str:
        """"""
        # super().__str__() produces:
        #     TypeError: super(type, obj): obj must be an instance or subtype of type
        # This is due to a combination of dataclass and slots=True.
        output = node_t.__str__(self)
        output += f"\n    Sites: {self.sites}\n    Diameters: {self.diameters}"
        return output


def EndNodes(
    tmap: array_t, /, *, width_map: array_t = None
) -> tuple[list[end_node_t], array_t]:
    """
    Note: End nodes are necessarily single-pixel nodes. Hence, they have no coordinate
    list.
    """
    # Not uint to allow for subtraction
    e_node_lmap = (tmap == 1).astype(nmpy.int64)  # Not really an lmsk here
    e_node_coords = nmpy.where(e_node_lmap)

    e_nodes: list[end_node_t] = e_node_coords[0].__len__() * [None]
    for n_idx, position in enumerate(zip(*e_node_coords)):
        end_node = end_node_t.NewWithPosition(
            nmpy.array(position, dtype=nmpy.int64), width_map=width_map
        )
        e_nodes[n_idx] = end_node
        e_node_lmap[position] = n_idx + 1  # Now that's becoming an lmsk

    return e_nodes, e_node_lmap


def BranchNodes(
    tmap: array_t, /, *, width_map: array_t = None
) -> tuple[list[branch_node_t], array_t]:
    """
    Note: Branch nodes always have a coordinate list (i.e., even if they are
    single-pixeled).
    """
    b_node_map = nmpy.logical_and(tmap > 2, tmap != tgmp.TMapBackgroundLabel(tmap))
    b_node_lmap, n_b_nodes = tgmp.LABELING_FCT_FOR_DIM[tmap.ndim](b_node_map)

    b_node_props = sims.regionprops(b_node_lmap)

    b_nodes: list[branch_node_t] = n_b_nodes * [None]
    for n_idx, props in enumerate(b_node_props):
        sites = props.image.nonzero()
        for d_idx in range(tmap.ndim):
            sites[d_idx].__iadd__(props.bbox[d_idx])
        branch_node = branch_node_t.NewWithSites(sites, width_map=width_map)
        b_nodes[n_idx] = branch_node

    return b_nodes, b_node_lmap



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
A map is a Numpy ndarray representing one or several object on a background. It has several variants. The ones concerned
here are:
- Binary   map (bmp): ndarray of type int8  where the object(s) are labeled 1 and the background is 0.
- Topology map (tmp): ndarray of type int8  where each object site (pixel, voxel...) has a value between 0
and 3**D - 1, where D is the number of dimensions of the array, and the background is 3**D. The values correspond to
the number of neighboring sites in the object with the weakest connectivity (8 in 2-D, 26 in 3-D...).

It is implicit that all the functions below require the map(s) to be valid (i.e., follow(s) the definitions above).
"""

from typing import Callable

import numpy as nmpy
import scipy.ndimage as spim

array_t = nmpy.ndarray
labeling_fct_t = Callable[[array_t], tuple[array_t, int]]


_SQUARE_3x3 = nmpy.ones((3, 3), dtype=nmpy.uint8)
_SQUARE_3x3x3 = nmpy.ones((3, 3, 3), dtype=nmpy.uint8)
_LABELING_FCT_2D: labeling_fct_t = lambda a_map: spim.label(
    a_map, structure=_SQUARE_3x3, output=nmpy.int64
)
_LABELING_FCT_3D: labeling_fct_t = lambda a_map: spim.label(
    a_map, structure=_SQUARE_3x3x3, output=nmpy.int64
)
LABELING_FCT_FOR_DIM = (None, None, _LABELING_FCT_2D, _LABELING_FCT_3D)


_FULL_SHIFTS_FOR_2D_NEIGHBORS = tuple(
    (i, j) for i in (-1, 0, 1) for j in (-1, 0, 1) if i != 0 or j != 0
)
_FULL_SHIFTS_FOR_3D_NEIGHBORS = tuple(
    (i, j, k)
    for i in (-1, 0, 1)
    for j in (-1, 0, 1)
    for k in (-1, 0, 1)
    if i != 0 or j != 0 or k != 0
)
_FULL_SHIFTS_FOR_NEIGHBORS_FOR_DIM = (
    None,
    None,
    _FULL_SHIFTS_FOR_2D_NEIGHBORS,
    _FULL_SHIFTS_FOR_3D_NEIGHBORS,
)

_MIN_SHIFTS_FOR_2D_NEIGHBORS = tuple(
    elm for elm in _FULL_SHIFTS_FOR_2D_NEIGHBORS if nmpy.abs(elm).sum() == 1
)
_MIN_SHIFTS_FOR_3D_NEIGHBORS = tuple(
    elm for elm in _FULL_SHIFTS_FOR_3D_NEIGHBORS if nmpy.abs(elm).sum() == 1
)
_MIN_SHIFTS_FOR_NEIGHBORS_FOR_DIM = (
    None,
    None,
    _MIN_SHIFTS_FOR_2D_NEIGHBORS,
    _MIN_SHIFTS_FOR_3D_NEIGHBORS,
)


def TopologyMapOfMap(
    a_map: array_t, /, *, full_connectivity: bool = True, return_bg_label: bool = False
) -> array_t | tuple[array_t, int]:
    """
    The topology map is labeled as follows: background=TMapBackgroundLabel(a_map); Pixels of the objects=number of
    neighboring pixels that belong to the given object (as expected, isolated pixels receive 0).

    Output dtype is int instead of uint to allow for subtraction

    Works for multi-object maps.

    Note: using a_map avoids shadowing Python's map.
    """
    output = nmpy.array(a_map, dtype=nmpy.int8)

    if full_connectivity:
        shifts_for_dim = _FULL_SHIFTS_FOR_NEIGHBORS_FOR_DIM
    else:
        shifts_for_dim = _MIN_SHIFTS_FOR_NEIGHBORS_FOR_DIM
    padded_sm = nmpy.pad(a_map, 1)
    unpadding_domain = a_map.ndim * (slice(1, -1),)
    rolling_axes = tuple(range(a_map.ndim))
    for shifts in shifts_for_dim[a_map.ndim]:
        output += nmpy.roll(padded_sm, shifts, axis=rolling_axes)[unpadding_domain]

    background_label = TMapBackgroundLabel(a_map)
    output[a_map == 0] = background_label + 1
    output -= 1

    if return_bg_label:
        return output, background_label

    return output


def TMapBackgroundLabel(a_map: array_t, /) -> int:
    """
    Must be equal to the max number of neighbors in a map, + 1.
    Note: using a_map avoids shadowing Python's map.
    """
    return 3**a_map.ndim



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Callable

import networkx as ntkx
import numpy as nmpy

from skl_graph.task.plot.base import (
    FigureAndAxesFromBoth,
    ManagePlotOptions,
    YTransformations,
    YTransformationsFromAxes,
    axes_3d_t,
    axes_any_h,
    axes_t,
    figure_t,
    y_transform_h,
)
from skl_graph.task.plot.edge import Plot as PlotEdges
from skl_graph.task.plot.node import (
    Plot2DBranchNodes,
    Plot2DEndNodes,
    Plot2DNodeLabels,
    Plot3DBranchNodes,
    Plot3DEndNodes,
    Plot3DNodeLabels,
    PositionsForPlotFromDetails,
    positions_as_dict_h,
)
from skl_graph.type.graph import skl_graph_t
from skl_graph.type.style import plot_mode_e

array_t = nmpy.ndarray


def Plot(
    graph: skl_graph_t,
    /,
    *,
    figure: figure_t = None,
    axes: axes_any_h = None,
    mode: plot_mode_e = plot_mode_e.SKL_Pixel,
    should_block: bool = True,
    should_return_figure: bool = False,
    should_return_axes: bool = False,
) -> figure_t | axes_t | axes_3d_t | tuple[figure_t, axes_any_h] | None:
    """"""
    if graph.n_nodes < 1:
        return None

    figure, axes = FigureAndAxesFromBoth(figure, axes, graph.dim)
    TransformedY, TransformedYForVector = YTransformationsFromAxes(
        axes, graph.domain_shape[0]
    )
    positions_as_dict = PositionsForPlotFromDetails(
        graph.nodes.data("details"), TransformedY
    )

    if mode is plot_mode_e.Networkx:
        if graph.dim != 2:
            raise ValueError(
                f'{graph.dim}: Invalid dimension for plot mode "plot_mode_e.Networkx". Expected=2.'
            )

        if graph.label_styles[1].show:
            edge_labels = EdgeLabelsForPlot(graph)
        else:
            edge_labels = None
        _PlotWithNetworkX(
            graph,
            axes,
            node_positions=positions_as_dict,
            node_colors=NodeColors(graph),
            with_node_labels=graph.label_styles[0].show,
            node_font_size=int(round(graph.label_styles[0].size)),
            edge_width=graph.edge_styles[0].size,
            edge_labels=edge_labels,
            edge_font_size=int(round(graph.label_styles[1].size)),
        )
    else:
        _PlotExplicitly(
            graph,
            positions_as_dict,
            TransformedY,
            TransformedYForVector,
            axes,
            mode,
        )

    return ManagePlotOptions(
        figure,
        axes,
        should_block,
        should_return_figure,
        should_return_axes,
    )


def PlotNetworkXGraph(
    graph: ntkx.MultiGraph,
    /,
    *,
    NodePositions: Callable[[axes_t], positions_as_dict_h] = None,
    node_colors: tuple[str, ...] = None,
    with_node_labels: bool = False,
    node_font_size: int = None,
    edge_width: int = None,
    edge_labels: dict[str, str] = None,
    edge_font_size: int = None,
    figure: figure_t = None,
    axes: axes_t = None,
    should_block: bool = True,
    should_return_figure: bool = False,
    should_return_axes: bool = False,
) -> figure_t | axes_t | axes_3d_t | tuple[figure_t, axes_t] | None:
    """
    Only 2-D graphs
    """
    if graph.number_of_nodes() < 1:
        return None

    figure, axes = FigureAndAxesFromBoth(figure, axes, 2)

    if NodePositions is None:
        node_positions = None
    else:
        node_positions = NodePositions(axes)
    _PlotWithNetworkX(
        graph,
        axes,
        node_positions=node_positions,
        node_colors=node_colors,
        with_node_labels=with_node_labels,
        node_font_size=node_font_size,
        edge_width=edge_width,
        edge_labels=edge_labels,
        edge_font_size=edge_font_size,
    )

    return ManagePlotOptions(
        figure,
        axes,
        should_block,
        should_return_figure,
        should_return_axes,
    )


def NodePositionsForPlot(
    graph: skl_graph_t, /
) -> Callable[[axes_any_h], positions_as_dict_h]:
    """"""
    YTransformationsForAxes = YTransformations(graph.domain_shape[0])

    return lambda _axs: PositionsForPlotFromDetails(
        graph.nodes.data("details"), YTransformationsForAxes(_axs)[0]
    )


def NodeColors(graph: skl_graph_t, /) -> tuple[str, ...]:
    """"""
    degrees = graph.degree
    node_styles = graph.node_styles
    default_style = node_styles[None]

    return tuple(
        node_styles.get(degrees[_nde], default_style).color for _nde in graph.nodes
    )


def EdgeLabelsForPlot(graph: skl_graph_t, /) -> dict[str, str]:
    """"""
    output = {}

    details_as_dict = ntkx.get_edge_attributes(graph, "details")
    for key, value in details_as_dict.items():
        output[key[0:2]] = f"{key[2]}\n{round(value.lengths.length)}"

    return output


def SetTightAxes(
    axes: axes_any_h, graph: skl_graph_t, /, *, with_aspect: bool = True
) -> None:
    """"""
    bbox = graph.bbox
    TransformedY, _ = YTransformationsFromAxes(axes, graph.domain_shape[0])

    axes.set_xlim(bbox[2], bbox[3])
    axes.set_ylim(TransformedY(bbox[0]), TransformedY(bbox[1]))
    if graph.dim == 3:
        axes.set_zlim(bbox[4], bbox[5])
        dim_indices = (2, 0, 4)
    else:
        dim_indices = (2, 0)
    if with_aspect:
        axes.set_box_aspect(tuple(bbox[_idx + 1] - bbox[_idx] for _idx in dim_indices))


def _PlotExplicitly(
    graph: skl_graph_t,
    node_label_positions: positions_as_dict_h,
    TransformedY: y_transform_h,
    TransformedYForVector: y_transform_h,
    axes: axes_any_h,
    mode: plot_mode_e,
    /,
) -> None:
    """"""
    PlotEdges(
        graph.edges.data("details", keys=True),
        TransformedY,
        TransformedYForVector,
        axes,
        graph.edge_styles,
        graph.direction_style,
        graph.label_styles[1],
        mode,
    )

    if graph.dim == 2:
        PlotEndNodes = Plot2DEndNodes
        PlotBranchNodes = Plot2DBranchNodes
    else:
        PlotEndNodes = Plot3DEndNodes
        PlotBranchNodes = Plot3DBranchNodes
    # Note: there is no distinction between end nodes and isolated nodes
    PlotEndNodes(
        graph.nodes.data("details"),
        TransformedY,
        axes,
        graph.node_styles,
    )
    PlotBranchNodes(
        graph.nodes.data("details"),
        graph.degree,
        TransformedY,
        axes,
        graph.node_styles,
    )

    if graph.label_styles[0].show:
        if graph.dim == 2:
            Plot2DNodeLabels(graph, node_label_positions, axes)
        else:
            Plot3DNodeLabels(graph, node_label_positions, axes, graph.label_styles[0])


def _PlotWithNetworkX(
    graph: ntkx.MultiGraph | skl_graph_t,
    axes: axes_t,
    /,
    *,
    node_positions: positions_as_dict_h = None,
    node_colors: tuple[str, ...] = None,
    with_node_labels: bool = False,
    node_font_size: int = None,
    edge_width: int = None,
    edge_labels: dict[str, str] = None,
    edge_font_size: int = None,
) -> None:
    """"""
    options = {}
    if node_positions is not None:
        options["pos"] = node_positions
    if node_colors is not None:
        options["node_color"] = node_colors
    if node_font_size is not None:
        options["font_size"] = node_font_size
    if edge_width is not None:
        options["width"] = edge_width
    ntkx.draw_networkx(
        graph,
        ax=axes,
        with_labels=with_node_labels,
        **options,
    )

    if (edge_labels is not None) and (node_positions is not None):
        options = {"edge_labels": edge_labels}
        if edge_font_size is not None:
            options["font_size"] = edge_font_size
        ntkx.draw_networkx_edge_labels(
            graph,
            node_positions,
            ax=axes,
            **options,
        )



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Any, Sequence

import numpy as nmpy

array_t = nmpy.ndarray
dtype_t = nmpy.dtype


def Issues(
    array: Any,
    /,
    *,
    expected_dtype: dtype_t | Sequence[dtype_t] = None,
    expected_size: int | Sequence[int] = None,
    expected_shape: Sequence[int] | Sequence[Sequence[int]],
) -> Sequence[str]:
    """"""
    output = []

    if not isinstance(array, array_t):
        return (f"{array}: Not a Numpy array. Actual type={type(array).__name__}.",)

    if (expected_dtype is not None) and not isinstance(expected_dtype, Sequence):
        expected_dtype = (expected_dtype,)
    if (expected_size is not None) and not isinstance(expected_size, Sequence):
        expected_size = (expected_size,)
    if (expected_shape is not None) and not isinstance(expected_shape[0], Sequence):
        expected_shape = (expected_shape,)

    if not any(
        (array.dtype == _tpe) or nmpy.issubdtype(array.dtype, _tpe)
        for _tpe in expected_dtype
    ):
        output.append(
            f"{array.dtype}: Invalid Numpy dtype. Expected={','.join(expected_dtype)}."
        )

    if array.size in expected_size:
        if array.shape not in expected_shape:
            output.append(
                f"{array.shape}: Invalid array shape. "
                f"Expected={','.join(expected_shape)}."
            )
    else:
        output.append(
            f"{array.size}: Invalid array size. Expected={','.join(expected_size)}."
        )

    return output



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

from typing import NamedTuple as named_tuple_t

import numpy as nmpy

array_t = nmpy.ndarray


class lengths_t(named_tuple_t):
    """
    segment_: prefix for pixel-to-pixel segments
    segment_sq_lengths: squared segment lengths; Interest: all integers
    """

    length: float
    segment_lengths: array_t
    segment_sq_lengths: array_t

    @classmethod
    def NewFromSites(cls, sites: tuple[array_t, ...], /) -> lengths_t:
        """"""
        sites_as_array = nmpy.array(sites)
        segments = nmpy.diff(sites_as_array, axis=1)
        segment_sq_lengths = nmpy.sum(segments**2, axis=0)
        segment_lengths = nmpy.sqrt(segment_sq_lengths)
        length = segment_lengths.sum().item()

        return cls(
            length=length,
            segment_lengths=segment_lengths,
            segment_sq_lengths=segment_sq_lengths,
        )


class areas_t(named_tuple_t):
    """
    segment_: prefix for pixel-to-pixel segments
    segment_based_area: width-weighted length
    """

    mean_width: float
    mean_based_area: float | None
    segment_based_area: float | None

    @classmethod
    def NewFromLengthsAndWidths(
        cls,
        lengths: lengths_t,
        widths: array_t,
        /,
    ) -> areas_t:
        """"""
        mean_width = nmpy.mean(widths).item()
        mean_based_area = lengths.length * mean_width
        segment_based_area = (
            (0.5 * (widths[1:] + widths[:-1]) * lengths.segment_lengths).sum().item()
        )

        return cls(
            mean_width=mean_width,
            mean_based_area=mean_based_area,
            segment_based_area=segment_based_area,
        )



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Optional

import numpy as nmpy

import skl_graph.type.tpy_map as tymp

array_t = nmpy.ndarray


def CheckSkeletonMap(
    skl_map: array_t,
    /,
    *,
    mode: Optional[str] = "single",
    behavior: Optional[str] = "exception",
) -> list[str] | None:
    """Raises an exception or returns a list of invalid properties if the passed map is not a valid skeleton map.

    The map dtype is not strictly checked: only floating point types raise an exception (but int64, for example, does
    not although the chosen definition for skeleton map only mentions boolean and 8-bit integer types). The other
    aspects of a valid skeleton map are described in the module documentation.

    Parameters
    ----------
    skl_map : numpy.ndarray
    mode : str, optional
        Can be "single" (the default) to check that `skl_map` is a valid skeleton map with a unique connected component,
        or "multi" if multiple connected components are allowed. It can also be None to skip validation.
    behavior : str, optional
        Can be "exception" (the default) to trigger an exception raising if the map is invalid, or "report" to just
        return None if the map is valid or a list of strings describing the invalid properties.

    Returns
    -------
    list[str], optional

    """
    # TODO: Fully check skl_map validity, like with node position, and do it first.
    if mode is None:
        return None

    if mode == "single":
        output = _SingleSkeletonMapIssues(skl_map)
    elif mode == "multi":
        output = _MultiSkeletonMapIssues(skl_map)
    else:
        raise ValueError(f'{mode}: Invalid "mode" value')

    if output is None:
        return None
    elif behavior == "exception":
        output = "\n    ".join(output)
        raise ValueError(f"Invalid {mode}-skeleton:\n    {output}")
    elif behavior == "report":
        return output
    else:
        raise ValueError(f'{behavior}: Invalid "behavior" value')


def _SingleSkeletonMapIssues(skl_map: array_t, /) -> list[str] | None:
    """Returns a list of invalid properties, if any, of the passed map when expecting a skeleton with a single connected
    component.

    Parameters
    ----------
    skl_map : numpy.ndarray

    Returns
    -------
    list[str], optional

    """
    output = _MultiSkeletonMapIssues(skl_map)
    if output is None:
        output = []

    _, n_components = tymp.LABELING_FCT_FOR_DIM[skl_map.ndim](skl_map)
    if n_components > 1:
        output.append(
            f"{n_components}: Too many connected components in map; Expected: 1"
        )

    if (output is None) or (output.__len__() == 0):
        return None
    return output


def _MultiSkeletonMapIssues(skl_map: array_t, /) -> list[str] | None:
    """Returns a list of invalid properties, if any, of the passed map when expecting a skeleton with one or more
    connected components.

    Parameters
    ----------
    skl_map : numpy.ndarray

    Returns
    -------
    list[str], optional

    """
    output = []

    if nmpy.issubdtype(skl_map.dtype, nmpy.floating):
        output.append(
            f"{skl_map.dtype}: Invalid map dtype; Expected: {nmpy.bool} or variants of {nmpy.integer}"
        )

    if skl_map.ndim not in (2, 3):
        output.append(f"{skl_map.ndim}: Invalid map dimension; Expected: 2 or 3")

    unique_values = nmpy.unique(skl_map)
    if not nmpy.array_equal(unique_values, (0, 1)):
        output.append(
            f"{unique_values}: Too many unique values in map; Expected: 0 and 1"
        )

    if (output is None) or (output.__len__() == 0):
        return None
    return output



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
Skeleton Map Creation and Manipulation.

Definitions
-----------
A map is a Numpy ndarray representing one or several object over a background. It exists in several variants:
- Boolean  map: ndarray of type bool            where the object(s) are labeled True and the background is False.
- Binary   map: ndarray of type int8  or uint8  where the object(s) are labeled 1 and the background is 0.
- Labeled  map: ndarray of type int64 or uint64 where the objects   are labeled from 1 with successive integers,
and the background is 0.
- Topology map: ndarray of type int8  or uint8  where each object site (pixel, voxel...) has a value between 0 and
3^D - 1, where D is the number of dimensions of the array, and the background is 3^D. The values correspond to the
number of neighboring sites belonging to the object with the weakest connectivity (8 in 2-D, 26 in 3-D...).

These variants are abbreviated omp, ymp, lmp, and tmp, respectively, to be used as a variable name postfix, e.g.,
edge_ymp.

For maps that can be of signed or unsigned types, the signed version is usually preferred to make them subtractable.

A skeleton map, normally abbreviated as skl_map, is a 2- or 3-dimensional boolean or binary map in which no site (pixel
or voxel) with value True, respectively 1, can be set to False, respectively 0, without breaking the skeleton
connectivity (in the weakest sense) or shortening a branch.

Simple example usage:
>>> # --- Object
>>> import skimage.data as data
>>> import skimage.util as util
>>> object_map = util.invert(data.horse())
>>> # --- SKL Map
>>> from skl_graph.type.skl_map import SKLMapFromObjectMap, PruneSKLMapBasedOnWidth
>>> skl_map, width_map = SKLMapFromObjectMap(object_map, with_width=True)
>>> pruned_map = skl_map.copy()
>>> PruneSKLMapBasedOnWidth(pruned_map, width_map, 20)
>>> # --- Plotting
>>> import matplotlib.pyplot as pyplot
>>> _, all_axes = pyplot.subplots(ncols=4)
>>> all_axes[0].matshow(object_map, cmap="gray")
>>> all_axes[1].matshow(skl_map, cmap="gray")
>>> all_axes[2].matshow(width_map, cmap="hot")
>>> all_axes[3].matshow(pruned_map, cmap="gray")
>>> for axes, title in zip(all_axes, ("Object", "Skeleton", "Width", "Pruned Skeleton")):
>>>     axes.set_title(title)
>>>     axes.set_axis_off()
>>> pyplot.tight_layout()
>>> pyplot.show()
"""

import numpy as nmpy
import scipy.ndimage as spim
import skimage.morphology as skmp

import skl_graph.type.tpy_map as tymp

array_t = nmpy.ndarray


def SKLMapFromObjectMap(
    object_map: array_t, /, *, with_width: bool = False
) -> array_t | tuple[array_t, array_t]:
    """Returns the skeleton map of an object map, optionally with the width map (see `SkeletonWidthMapFromObjectMap`).

    Works for multiple objects if skmp.thin and skmp.skeletonize_3d do.

    Parameters
    ----------
    object_map : numpy.ndarray
    with_width : bool

    Returns
    -------
    array_t | tuple[array_t, array_t]

    """
    if object_map.ndim == 2:
        # Documentation says it removes every pixel up to breaking connectivity
        Skeletonized = skmp.thin
    elif object_map.ndim == 3:
        # Documentation does not tell anything about every pixel being necessary or not
        Skeletonized = skmp.skeletonize_3d
    else:
        raise ValueError(f"{object_map.ndim}: Invalid map dimension; Expected: 2 or 3")

    skl_map = Skeletonized(object_map).astype(nmpy.bool_, copy=False)
    if object_map.ndim == 3:
        SKLMapFromThickVersion(skl_map, in_place=True)

    if with_width:
        return skl_map, SkeletonWidthMapFromObjectMap(object_map)
    return skl_map


def SkeletonWidthMapFromObjectMap(object_map: array_t, /) -> array_t:
    """Width map of an object map.

    The width map is a distance map where the values on the object(s) skeleton are equal to twice the distance to the
    object border, which can be interpreted as the local object width.

    Parameters
    ----------
    object_map : numpy.ndarray

    Returns
    -------
    numpy.ndarray

    """
    return 2.0 * spim.distance_transform_edt(object_map) + 1.0


_CENTER_3x3 = ((0, 0, 0), (0, 1, 0), (0, 0, 0))
_CROSS_3x3 = nmpy.array(((0, 1, 0), (1, 1, 1), (0, 1, 0)), dtype=nmpy.uint8)
_CROSS_3x3x3 = nmpy.array((_CENTER_3x3, _CROSS_3x3, _CENTER_3x3), dtype=nmpy.uint8)
_CROSS_FOR_DIM = (None, None, _CROSS_3x3, _CROSS_3x3x3)


def SKLMapFromThickVersion(
    skl_map: array_t, /, *, in_place: bool = False, should_only_check: bool = False
) -> array_t | bool | None:
    """Removes all sites (pixels or voxels) that do not break the skeleton connectivity (in the weakest sense) or
    shorten a branch.

    Works for multi-skeletons.
    """
    dtype = skl_map.dtype
    is_boolean = nmpy.issubdtype(dtype, nmpy.bool_)
    if not (is_boolean or nmpy.issubdtype(dtype, nmpy.integer)):
        raise ValueError(
            f"{dtype.name}: Invalid Numpy dtype. Expected=bool_ or integer-like."
        )

    if is_boolean:
        min_value = False
        max_value = True
    else:
        unique_values = nmpy.unique(skl_map)
        if (unique_values[0] != 0) or (unique_values.size != 2):
            raise ValueError(
                f"{unique_values}: Invalid unique values. Expected=0 and a strictly positive value."
            )
        min_value, max_value = unique_values

    skl_dimension = skl_map.ndim
    background_label = tymp.TMapBackgroundLabel(skl_map)
    patch_center = skl_dimension * (1,)
    cross = _CROSS_FOR_DIM[skl_dimension]
    padded_thinned = nmpy.pad(skl_map, 1)
    LabeledMap = tymp.LABELING_FCT_FOR_DIM[skl_dimension]

    def _DoFixPatches(
        _tpy_map: array_t,
        _n_neighbors: int,
        /,
    ) -> bool:
        """"""
        _skel_has_been_modified = False

        for where in zip(*nmpy.where(_tpy_map == _n_neighbors)):
            patch_slices = tuple(slice(coord - 1, coord + 2) for coord in where)
            tpy_patch = _tpy_map[patch_slices]
            if nmpy.any(tpy_patch[cross] == background_label):
                patch = padded_thinned[patch_slices]
                patch[patch_center] = min_value

                _, n_components = LabeledMap(patch)
                if n_components == 1:
                    _skel_has_been_modified = True
                else:
                    patch[patch_center] = max_value

        return _skel_has_been_modified

    def _JustCheckPatches(
        _tpy_map: array_t,
        _n_neighbors: int,
        /,
    ) -> bool:
        """"""
        for where in zip(*nmpy.where(_tpy_map == _n_neighbors)):
            patch_slices = tuple(slice(coord - 1, coord + 2) for coord in where)
            tpy_patch = _tpy_map[patch_slices]
            if nmpy.any(tpy_patch[cross] == background_label):
                patch = padded_thinned[patch_slices]
                patch[patch_center] = min_value

                _, n_components = LabeledMap(patch)
                if n_components == 1:
                    return True
                else:
                    patch[patch_center] = max_value

        return False

    if should_only_check:
        _FixPatches = _JustCheckPatches
    else:
        _FixPatches = _DoFixPatches
    excluded_n_neighbors = {
        0,
        1,
        2 * skl_dimension,
        background_label,
    }
    skel_has_been_modified = True
    while skel_has_been_modified:
        skel_has_been_modified = False

        tpy_map = tymp.TopologyMapOfMap(padded_thinned, full_connectivity=False)
        included_n_neighbors = set(nmpy.unique(tpy_map)).difference(
            excluded_n_neighbors
        )

        for n_neighbors in sorted(included_n_neighbors, reverse=True):
            skel_has_been_modified = skel_has_been_modified or _FixPatches(
                tpy_map,
                n_neighbors,
            )
            if should_only_check and skel_has_been_modified:
                return False

    if should_only_check:
        return True

    if skl_dimension == 2:
        thinned = padded_thinned[1:-1, 1:-1]
    else:
        thinned = padded_thinned[1:-1, 1:-1, 1:-1]

    if in_place:
        nmpy.copyto(skl_map, thinned)
    else:
        return thinned


def PruneSKLMapBasedOnWidth(
    skl_map: array_t, width_map: array_t, min_width: float, /
) -> None:
    """Prunes the skeleton map so that the resulting skeleton corresponds everywhere to object portions wider than the
    passed minimal width.

    Works for multi-skeletons.

    Parameters
    ----------
    skl_map : numpy.ndarray
    width_map : numpy.ndarray
    min_width : float

    Returns
    -------
    None

    """
    while True:
        tpy_map = tymp.TopologyMapOfMap(skl_map)
        end_positions = nmpy.where(tpy_map == 1)
        distances = width_map[end_positions]

        tiny_distances = distances < min_width
        if tiny_distances.any():
            extra_positions = tuple(site[tiny_distances] for site in end_positions)
            skl_map[extra_positions] = 0
        else:
            break



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Callable

import numpy as nmpy
from matplotlib import pyplot as pypl
from mpl_toolkits.mplot3d import Axes3D as axes_3d_t

array_t = nmpy.ndarray

axes_t = pypl.Axes
axes_any_h = axes_t | axes_3d_t
figure_t = pypl.Figure

NewFigure = pypl.figure
EnterMatplotlibEventLoop = pypl.show

y_transform_h = Callable[[int | array_t], array_t]


def FigureAndAxesFromBoth(
    figure: figure_t | None, axes: axes_t | axes_3d_t | None, dimension: int, /
) -> tuple[figure_t, axes_any_h]:
    """"""
    if axes is None:
        if figure is None:
            figure = NewFigure()
        if dimension == 2:
            axes = figure.gca()
        else:
            axes = figure.add_subplot(1, 1, 1, projection=axes_3d_t.name)
        axes.invert_yaxis()
    else:
        figure = axes.get_figure()

    return figure, axes


def YTransformationsFromAxes(
    axes: axes_any_h, domain_height: int, /
) -> tuple[y_transform_h, y_transform_h]:
    """
    Note: For sites, an inverted y-axis (increasing to the bottom) is the normal, row/col convention for SKL-Graph, so
    TransformedY is the identity. However, it would flip vectors, so TransformedYForVector should indeed flip them.
    """
    # nmpy.asarray: To be able to transform several y's at once
    if axes.yaxis_inverted():
        TransformedY = lambda y: y
        TransformedYForVector = lambda y: -nmpy.asarray(y)
    else:
        max_0 = domain_height - 1
        TransformedY = lambda y: max_0 - nmpy.asarray(y)
        TransformedYForVector = lambda y: y

    return TransformedY, TransformedYForVector


def YTransformations(
    domain_height: int, /
) -> Callable[[axes_any_h], tuple[y_transform_h, y_transform_h]]:
    """"""
    return lambda _axs: YTransformationsFromAxes(_axs, domain_height)


def ManagePlotOptions(
    figure: figure_t,
    axes: axes_any_h,
    should_block: bool,
    should_return_figure: bool,
    should_return_axes: bool,
    /,
) -> figure_t | axes_t | axes_3d_t | tuple[figure_t, axes_any_h] | None:
    """"""
    # Preferably test not-axes_3d_t instead of yep-axes_t in case axes_3d_t is a subclass of axes_t
    if not isinstance(axes, axes_3d_t):
        # Matplotlib says: NotImplementedError: It is not currently possible to manually set the aspect on 3D axes
        axes.axis("equal")

    if should_block:
        EnterMatplotlibEventLoop()
        return None
    elif should_return_figure:
        if should_return_axes:
            return figure, axes
        else:
            return figure
    elif should_return_axes:
        return axes

    return None



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from typing import Any, Sequence

import numpy as nmpy
import scipy.interpolate as spnt
import skimage.measure as sims

import skl_graph.type.tpy_map as tgmp
from skl_graph.extension.number import COORDINATE_SEPARATOR, EncodedNumber
from skl_graph.type.feature import areas_t, lengths_t
from skl_graph.type.node import branch_node_t, end_node_t, node_t

array_t = nmpy.ndarray


_MAX_SQ_INTERPOLATION_ERROR = 0.8**2


@dtcl.dataclass(slots=True, repr=False, eq=False)
class raw_edge_t:
    """
    /!\ "sites" is organized by dimension, not by site. This is why "dim" is defined as self.sites.__len__() instead of
    self.sites[0].size.
    """

    sites: tuple[array_t, ...]

    @property
    def dim(self) -> int:
        """"""
        return self.sites.__len__()

    @property
    def n_sites(self) -> int:
        """
        Cannot be less than 2 since even an edge between an end node and a touching branch node will contain the end
        node pixel and one branch node pixel.
        """
        return self.sites[0].size


@dtcl.dataclass(slots=True, repr=False, eq=False)
class edge_t(raw_edge_t):
    uid: str = None
    lengths: lengths_t = None
    widths: array_t = None
    areas: areas_t = None
    properties: dict[Any, Any] = dtcl.field(init=False, default_factory=dict)
    _cache: dict[str, Any] = dtcl.field(default_factory=dict)

    @classmethod
    def NewWithDetails(
        cls,
        sites: tuple[array_t, ...],
        sites_are_ordered: bool,
        adjacent_node_uids: Sequence[str],
        /,
        *,
        width_map: array_t = None,
    ) -> edge_t:
        """"""
        if sites_are_ordered:
            ordered_sites = sites
        else:
            ordered_sites = _ReOrderedSites(sites)

        output = cls(ordered_sites)

        output.uid = _UID(adjacent_node_uids)
        output.lengths = lengths_t.NewFromSites(ordered_sites)
        if width_map is not None:
            output.widths = width_map[ordered_sites]
            output.areas = areas_t.NewFromLengthsAndWidths(
                output.lengths, output.widths
            )

        return output

    def SetProperty(self, name: str, /, *, value: Any = None):
        """"""
        self.properties[name] = value

    @property
    def arc_lengths(self) -> array_t:
        """
        Arc length coordinates of each pixel/voxel composing the edge.
        """
        output = nmpy.cumsum((0.0,) + tuple(self.lengths.segment_lengths))
        output[-1] = self.lengths.length  # To correct round-off errors.
        assert output[-1] > output[-2]  # To check above correction was not invalid.

        return output

    def AsCurve(self) -> spnt.PchipInterpolator | None:
        """"""
        cache_entry = self.AsCurve.__name__

        if cache_entry not in self._cache:
            if self.n_sites > 3:
                self._cache[cache_entry] = _PchipInterpolator(
                    self.arc_lengths, nmpy.array(self.sites)
                )
            else:
                self._cache[cache_entry] = None

        return self._cache[cache_entry]

    def __str__(self) -> str:
        """
        Does not call raw_edge_t.__str__ since raw edges do not have UID
        """
        origin = tuple(self.sites[_idx][0] for _idx in range(self.dim))
        raw_length = round(self.lengths.length, 2)
        if self.areas is None:
            segment_based_area = raw_length
        else:
            segment_based_area = round(self.areas.segment_based_area, 2)

        main = (
            f"{self.__class__.__name__}.{self.uid}:\n"
            f"    Sites[{self.dim}-D]: {self.n_sites}\n"
            f"    Origin:              {origin}\n"
            f"    Lengths:             Raw={raw_length}, SB-Area={segment_based_area}"
        )

        if (self._cache is None) or (self._cache.__len__() == 0):
            cached_values = "None yet"
        else:
            cached_values = ", ".join(self._cache.keys())

        return main + f"\n    Cached values: {cached_values}"


@dtcl.dataclass(slots=True, repr=False, eq=False)
class _transient_t(raw_edge_t):
    def __post_init__(self) -> None:
        """"""
        self.sites = _ReOrderedSites(self.sites)

    def AppendBranchNode(
        self,
        b_coords: array_t,
        node: node_t,
        adjacent_node_uids: list[str],
        /,
        *,
        force_after: bool = False,
    ) -> None:
        """"""
        adjacent_node_uids.append(node.uid)

        space_dim = self.dim
        first_site = tuple(self.sites[idx_][0] for idx_ in range(space_dim))
        sq_distance = (nmpy.subtract(first_site, b_coords) ** 2).sum()

        if self.n_sites > 1:
            # 0 <: so that if the edge is a self-loop ending at the same site, it does not put twice the site in a row
            if 0 < sq_distance <= space_dim:
                self.sites = tuple(
                    nmpy.hstack((b_coords[idx_], self.sites[idx_]))
                    for idx_ in range(space_dim)
                )
            else:
                self.sites = tuple(
                    nmpy.hstack((self.sites[idx_], b_coords[idx_]))
                    for idx_ in range(space_dim)
                )
        elif force_after:
            self.sites = tuple(
                nmpy.hstack((self.sites[idx_], b_coords[idx_]))
                for idx_ in range(space_dim)
            )
        else:
            self.sites = tuple(
                nmpy.hstack((b_coords[idx_], self.sites[idx_]))
                for idx_ in range(space_dim)
            )


def RawEdges(
    skl_map: array_t, b_node_lmap: array_t, /
) -> tuple[Sequence[raw_edge_t], array_t]:
    """"""
    edge_map = skl_map.astype(nmpy.int8)
    edge_map[b_node_lmap > 0] = 0
    edge_lmap, n_edges = tgmp.LABELING_FCT_FOR_DIM[skl_map.ndim](edge_map)

    edge_props = sims.regionprops(edge_lmap)

    edges: list[raw_edge_t] = n_edges * [None]
    for props in edge_props:
        sites = props.image.nonzero()
        for d_idx in range(skl_map.ndim):
            sites[d_idx].__iadd__(props.bbox[d_idx])
        edges[props.label - 1] = raw_edge_t(sites)

    return edges, edge_lmap


def EdgesFromRawEdges(
    raw_edges: Sequence[raw_edge_t],
    e_nodes: Sequence[end_node_t],
    b_nodes: Sequence[branch_node_t],
    edge_lmap: array_t,
    e_node_lmap: array_t,
    b_node_lmap: array_t,
    /,
    *,
    width_map: array_t = None,
) -> tuple[tuple[edge_t], list[list[str]]]:
    """"""
    transient_edges = tuple(map(lambda _raw: _transient_t(_raw.sites), raw_edges))
    edge_tmap = tgmp.TopologyMapOfMap(edge_lmap > 0)

    # ep=edge end point; Keep < 2 since ==0 (length-1 edges) and ==1 (other edges) are needed
    # Do not use list multiplication since the same list then used for all the elements
    node_uids_per_edge: list[list[str]] = [[] for _ in transient_edges]
    for ep_coords in zip(*(edge_tmap < 2).nonzero()):
        edge_idx = edge_lmap[ep_coords] - 1
        transient = transient_edges[edge_idx]
        e_node_label = e_node_lmap[ep_coords]

        if e_node_label > 0:
            # End node-to-X edge (i.e., edge end point is also an end node)
            node_uids_per_edge[edge_idx].append(e_nodes[e_node_label - 1].uid)
            if transient.n_sites == 1:
                # End node-to-branch node edge (and there is a unique non-zero value in b_neighborhood)
                nh_slices_starts, b_neighborhood = _LMapNeighborhood(
                    b_node_lmap, ep_coords
                )
                b_node_label = nmpy.amax(b_neighborhood)
                b_coords = nmpy.transpose((b_neighborhood == b_node_label).nonzero())[0]
                transient.AppendBranchNode(
                    nmpy.add(nh_slices_starts, b_coords),
                    b_nodes[b_node_label - 1],
                    node_uids_per_edge[edge_idx],
                )
        else:
            nh_slices_starts, b_neighborhood = _LMapNeighborhood(b_node_lmap, ep_coords)
            force_after = False
            # Looping only for length-1, b-to-b edges
            for b_coords in zip(*b_neighborhood.nonzero()):
                b_node_label = b_neighborhood[b_coords]
                transient.AppendBranchNode(
                    nmpy.add(nh_slices_starts, b_coords),
                    b_nodes[b_node_label - 1],
                    node_uids_per_edge[edge_idx],
                    force_after=force_after,
                )
                force_after = not force_after

    edges = []
    for transient, adjacent_node_uids in zip(transient_edges, node_uids_per_edge):
        # Fix sites order for self-loops connected to branch nodes
        if adjacent_node_uids[0] == adjacent_node_uids[1]:
            node = None
            for current in b_nodes:
                if current.uid == adjacent_node_uids[0]:
                    node = current
                    break
            if node is not None:
                node_sites = nmpy.array(node.sites)
                edge_sites = nmpy.array(transient.sites)
                while not all(
                    any(
                        nmpy.all(
                            (node_sites - edge_sites[:, _sdx][:, None]) == 0, axis=0
                        )
                    )
                    for _sdx in (0, -1)
                ):
                    edge_sites = nmpy.roll(edge_sites, -1, axis=1)

                transient.sites = tuple(
                    edge_sites[_cdx, :] for _cdx in range(transient.dim)
                )

        edge = edge_t.NewWithDetails(
            transient.sites, True, adjacent_node_uids, width_map=width_map
        )
        edges.append(edge)

    return tuple(edges), node_uids_per_edge


def _ReOrderedSites(sites: tuple[array_t, ...], /) -> tuple[array_t, ...]:
    """
    If the number of sites is 1 or 2, the input argument is returned (i.e., no copy is made).
    """
    n_sites = sites[0].size
    if n_sites < 3:
        return sites

    dim = sites.__len__()

    self_loop = all(sites[idx][0] == sites[idx][-1] for idx in range(dim))
    if self_loop:
        sites = tuple(sites[idx][:-1] for idx in range(dim))
        n_sites -= 1
        self_origin = nmpy.fromiter(
            (sites[idx][0] for idx in range(dim)), dtype=sites[0].dtype
        )
        self_origin = nmpy.reshape(self_origin, (1, dim))
    else:
        self_origin = None

    sites_as_array = nmpy.transpose(nmpy.array(sites))
    reordered_coords = [nmpy.array([sites[idx][0] for idx in range(sites.__len__())])]
    unvisited_slc = nmpy.ones(n_sites, dtype=nmpy.bool_)
    unvisited_slc[0] = False
    unvisited_sites = None
    end_point = None
    pre_done = False
    post_done = False

    while unvisited_slc.any():
        if post_done:
            neighbor_idc = ()
        else:
            end_point = reordered_coords[-1]
            neighbor_idc, unvisited_sites = _NeighborIndices(
                dim, sites_as_array, unvisited_slc, end_point
            )

        if (neighbor_idc.__len__() == 1) or post_done:
            also_grow_first = (reordered_coords.__len__() > 1) and not pre_done
            if not post_done:
                c_idx = neighbor_idc[0]
                reordered_coords.append(unvisited_sites[c_idx, :])
                unvisited_slc[nmpy.where(unvisited_slc)[0][c_idx]] = False
            if also_grow_first:
                end_point = reordered_coords[0]
                neighbor_idc, unvisited_sites = _NeighborIndices(
                    dim, sites_as_array, unvisited_slc, end_point
                )
                if neighbor_idc.__len__() == 1:
                    c_idx = neighbor_idc[0]
                    reordered_coords = [unvisited_sites[c_idx, :]] + reordered_coords
                    unvisited_slc[nmpy.where(unvisited_slc)[0][c_idx]] = False
                elif neighbor_idc.__len__() == 0:
                    pre_done = True  # End point has been reached
                else:
                    raise RuntimeError(
                        f"{neighbor_idc.__len__()} neighbors when only 1 is expected\n"
                        f"{sites}\n{reordered_coords}\n{unvisited_slc}\n{end_point}"
                    )
        elif neighbor_idc.__len__() == 2:
            if reordered_coords.__len__() == 1:
                idx1, idx2 = neighbor_idc
                reordered_coords = [unvisited_sites[idx1, :]] + reordered_coords
                reordered_coords.append(unvisited_sites[idx2, :])
                true_map = nmpy.where(unvisited_slc)[0]
                unvisited_slc[true_map[idx1]] = False
                unvisited_slc[true_map[idx2]] = False
            else:
                raise RuntimeError(
                    f"2 neighbors when only 1 is expected\n"
                    f"{sites}\n{reordered_coords}\n{unvisited_slc}\n{end_point}"
                )
        elif neighbor_idc.__len__() == 0:
            post_done = True  # End point has been reached
        else:
            raise RuntimeError(
                f"{neighbor_idc.__len__()} neighbors when only 1 or 2 are expected\n"
                f"{sites}\n{reordered_coords}\n{unvisited_slc}\n{end_point}"
            )

    reordered_coords = nmpy.array(reordered_coords)
    if self_loop:
        reordered_coords = _RolledSitesWithFixedOrigin(reordered_coords, self_origin)
    reordered_coords = tuple(reordered_coords[:, _idx] for _idx in range(dim))

    return reordered_coords


def _NeighborIndices(
    dim: int, sites: array_t, unvisited_slc: array_t, end_point: array_t
) -> tuple[array_t, array_t]:
    """"""
    unvisited_sites = sites[unvisited_slc, :]

    distances = nmpy.fabs(unvisited_sites - nmpy.reshape(end_point, (1, dim)))
    neighbor_idc = nmpy.nonzero(nmpy.all(distances <= 1, axis=1))[0]

    return neighbor_idc, unvisited_sites


def _RolledSitesWithFixedOrigin(sites: array_t, origin: array_t, /) -> array_t:
    """"""
    self_origin_idx = nmpy.argwhere(nmpy.all(sites == origin, axis=1)).item()
    if self_origin_idx > 0:
        sites = nmpy.roll(sites, -self_origin_idx, axis=0)

    return nmpy.vstack((sites, origin))


def _LMapNeighborhood(lmap: array_t, site: tuple[int, ...]) -> tuple[array_t, array_t]:
    """"""
    slices_starts = tuple(max(site[idx_] - 1, 0) for idx_ in range(site.__len__()))
    slices = tuple(
        slice(slices_starts[idx_], min(site[idx_] + 2, lmap.shape[idx_]))
        for idx_ in range(site.__len__())
    )
    neighborhood = lmap[slices]

    return nmpy.array(slices_starts, dtype=nmpy.int64), neighborhood


def _PchipInterpolator(
    arc_lengths: array_t, sites: array_t, /
) -> spnt.PchipInterpolator:
    """"""
    for n_samples in range(2, arc_lengths.size + 1):
        arc_length_samples = nmpy.linspace(0, arc_lengths.size - 1, num=n_samples)
        indices = nmpy.unique(nmpy.around(arc_length_samples)).astype(nmpy.uint64)
        output = spnt.PchipInterpolator(
            arc_lengths[indices], sites[:, indices], axis=1, extrapolate=False
        )

        error = max(nmpy.sum((output(arc_lengths) - sites) ** 2, axis=0))
        if error <= _MAX_SQ_INTERPOLATION_ERROR:
            return output

    raise RuntimeError("Should never happen. Please contact developer.")


def _UID(adjacent_node_uids: Sequence[str], /) -> str:
    """"""
    if adjacent_node_uids.__len__() != 2:
        raise RuntimeError(
            f"{adjacent_node_uids.__len__()}: Incorrect number of adjacent node uids"
        )

    node_uid_0, node_uid_1 = adjacent_node_uids
    if node_uid_0 > node_uid_1:
        node_uid_0, node_uid_1 = node_uid_1, node_uid_0

    uid_components = [
        EncodedNumber(coord) for coord in node_uid_0.split(COORDINATE_SEPARATOR)
    ]
    uid_components.append(COORDINATE_SEPARATOR)
    uid_components.extend(
        EncodedNumber(coord) for coord in node_uid_1.split(COORDINATE_SEPARATOR)
    )

    return "".join(uid_components)



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Iterable

import numpy as nmpy

from skl_graph.feature.edge import InwardEndTangents
from skl_graph.task.plot.base import axes_t, y_transform_h
from skl_graph.type.edge import edge_t
from skl_graph.type.style import (
    direction_style_t,
    edge_styles_h,
    label_style_t,
    plot_mode_e,
)

_SUBSAMPLING_FACTOR = 0.75
_LABEL_BBOX_STYLE = {
    "facecolor": "w",
    "linestyle": "",
    "alpha": 0.5,
    "boxstyle": "square,pad=0",
}


def Plot(
    edges: Iterable[tuple[str, str, str, edge_t]],
    TransformedY: y_transform_h,
    TransformedYForVector: y_transform_h,
    axes: axes_t,
    edge_styles: edge_styles_h,
    direction_style: direction_style_t,
    label_style: label_style_t,
    mode: plot_mode_e,
    /,
) -> None:
    """"""
    # space_dim = edges[0][2].dim  # Does not work since 'MultiEdgeDataView' object is not subscriptable
    space_dim = 2
    for *_, details in edges:
        space_dim = details.dim
        break

    PlotEdge = axes.plot if space_dim == 2 else axes.plot3D
    PlotEdgeLabel = axes.text

    for source, target, uid, details in edges:
        if mode is plot_mode_e.SKL_Curve:
            as_curve = details.AsCurve()
            if as_curve is None:
                sites = list(details.sites)
            else:
                max_arc_length = as_curve.x.item(-1)
                n_samples = int(round(_SUBSAMPLING_FACTOR * details.lengths.length))
                arc_lengths = nmpy.linspace(0.0, max_arc_length, num=max(n_samples, 2))
                sites = [_row for _row in as_curve(arc_lengths)]
        elif mode is plot_mode_e.SKL_Pixel:
            sites = list(details.sites)
        else:
            raise ValueError(
                f"{mode}: Invalid plotting mode; Valid modes: plot_mode_e.SKL_*"
            )
        sites[0], sites[1] = sites[1], TransformedY(sites[0])

        if source == target:
            edge_style = edge_styles[1]
        else:
            edge_style = edge_styles[0]
        PlotEdge(
            *sites,
            edge_style.color + edge_style.type,
            linewidth=edge_style.size,
            markersize=edge_style.size,
        )

        if direction_style.show:
            dir_sites = tuple(
                nmpy.hstack((sites[idx_][0], sites[idx_][-1]))
                for idx_ in range(space_dim)
            )
            tangents = list(zip(*InwardEndTangents(details)))
            tangents[0], tangents[1] = (
                tangents[1],
                TransformedYForVector(tangents[0]),
            )
            axes.quiver(
                *dir_sites,
                *tangents,
                color=direction_style.color,
                linewidth=direction_style.size,
            )

        if label_style.show:
            middle_idx = sites[0].size // 2
            if sites[0].size % 2 == 0:
                position = tuple(
                    nmpy.mean(_crd[(middle_idx - 1) : (middle_idx + 1)]).item()
                    for _crd in sites
                )
            else:
                position = tuple(_crd[middle_idx] for _crd in sites)
            label = PlotEdgeLabel(
                *position,
                uid,
                fontsize=label_style.size,
                color=label_style.color,
                horizontalalignment="center",
                verticalalignment="center",
            )
            label.set_bbox(_LABEL_BBOX_STYLE)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import numpy as nmpy

from skl_graph.type.edge import edge_t

array_t = nmpy.ndarray


def InwardEndTangents(edge: edge_t, /) -> tuple[array_t, array_t]:
    """"""
    cache_entry = InwardEndTangents.__name__

    if cache_entry not in edge._cache:
        first_inner, last_inner = _InwardEndSegments(edge)
        flips = [1.0, 1.0]

        as_curve = edge.AsCurve()
        if as_curve is None:
            if edge.n_sites > 1:
                first_tangent = nmpy.array(first_inner, dtype=nmpy.float64)
                last_tangent = nmpy.array(last_inner, dtype=nmpy.float64)
            else:
                raise RuntimeError("Should never happen. Please contact developer.")
        else:
            max_arc_length = as_curve.x.item(-1)
            derivatives = as_curve((0, max_arc_length), 1)
            first_tangent = nmpy.array(derivatives[:, 0], dtype=nmpy.float64)
            last_tangent = nmpy.array(derivatives[:, 1], dtype=nmpy.float64)
            for p_idx, product in enumerate(
                (
                    nmpy.dot(first_tangent, first_inner),
                    nmpy.dot(last_tangent, last_inner),
                )
            ):
                if product < 0.0:
                    flips[p_idx] = -1.0
        first_tangent /= flips[0] * nmpy.linalg.norm(first_tangent)
        last_tangent /= flips[1] * nmpy.linalg.norm(last_tangent)
        edge._cache[cache_entry] = (first_tangent, last_tangent)

    return edge._cache[cache_entry]


def Tangents(edge: edge_t, /) -> tuple[array_t, array_t, array_t]:
    """"""
    cache_entry = Tangents.__name__

    if cache_entry not in edge._cache:
        as_curve = edge.AsCurve()
        if as_curve is None:
            if edge.n_sites > 1:
                sites = nmpy.array(edge.sites)
                arc_lengths = edge.arc_lengths
                derivatives = nmpy.diff(sites, axis=1).astype(nmpy.float64)
                norms = nmpy.linalg.norm(derivatives, axis=0, keepdims=True)
                derivatives /= norms
                last = derivatives[:, -1][:, None]
                tangents = nmpy.hstack((derivatives, last))
            else:
                raise RuntimeError("Should never happen. Please contact developer.")
        else:
            arc_lengths = as_curve.x
            sites, derivatives = (as_curve(arc_lengths, _rdr) for _rdr in (0, 1))
            norms = nmpy.linalg.norm(derivatives, axis=0, keepdims=True)
            tangents = derivatives / norms
        edge._cache[cache_entry] = (sites, arc_lengths, tangents)

    return edge._cache[cache_entry]


def TangentDirectionChanges(edge: edge_t, /) -> array_t:
    """
    2-D: determinant of consecutive tangents, normalized by arc length differences.
    3-D: "signed" norm of the cross product of consecutive tangents, normalized by arc length differences.
    """
    cache_entry = TangentDirectionChanges.__name__

    if cache_entry not in edge._cache:
        if edge.dim == 2:
            Determinant = lambda _vct, _ref: (nmpy.linalg.det(_vct), _ref)
        else:
            Determinant = _SignedCrossProductNorm

        _, arc_lengths, tangents = Tangents(edge)
        determinants = nmpy.empty((tangents.shape[1] - 1), dtype=nmpy.float64)
        reference = None
        for c_idx in range(tangents.shape[1] - 1):
            determinants[c_idx], reference = Determinant(
                tangents[:, c_idx : (c_idx + 2)], reference
            )
        determinants /= arc_lengths[1:]
        edge._cache[cache_entry] = determinants

    return edge._cache[cache_entry]


def Tortuosity(edge: edge_t, /) -> float:
    """
    https://en.wikipedia.org/wiki/Tortuosity
    """
    chord_length = nmpy.linalg.norm(edge.sites[0] - edge.sites[-1])
    return edge.lengths.length / chord_length


def _InwardEndSegments(edge: edge_t, /) -> tuple[tuple[int, ...], tuple[int, ...]]:
    """"""
    dim_range = tuple(range(edge.dim))
    first_inner = tuple(edge.sites[_idx][1] - edge.sites[_idx][0] for _idx in dim_range)
    last_inner = tuple(
        edge.sites[_idx][-2] - edge.sites[_idx][-1] for _idx in dim_range
    )

    return first_inner, last_inner


def _SignedCrossProductNorm(
    vectors: array_t, reference: array_t | None, /
) -> tuple[float, array_t]:
    """"""
    product = nmpy.cross(vectors[:, 0], vectors[:, 1])
    if reference is None:
        reference = product
        sign = 1.0
    else:
        sign = nmpy.dot(product, reference)

    return (sign * nmpy.linalg.norm(product)).item(), reference



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
>>> import numpy as nmpy
>>> from documentation.wiki.example.minimal import skl_graph, skl_map
>>> from json_any import JsonStringOf, ObjectFromJsonString
>>> from skl_graph.task.storage import DescriptionForJSON, NewFromJsonDescription
>>> from skl_graph.type.graph import skl_graph_t
>>>
>>> jsoned, history = JsonStringOf(skl_graph, descriptors={"": DescriptionForJSON})
>>> print(history)
None
>>> un_jsoned = ObjectFromJsonString(jsoned, builders={"": NewFromJsonDescription})
>>> print(type(un_jsoned).__name__)
skl_graph_t
>>> (un_jsoned.n_nodes == skl_graph.n_nodes) and (un_jsoned.n_edges == skl_graph.n_edges)
True
>>> nmpy.array_equal(un_jsoned.RebuiltSkeletonMap().astype(skl_map.dtype), skl_map)
True
"""

from typing import Any

import networkx as ntkx

from skl_graph.type.graph import skl_graph_t


def DescriptionForJSON(graph: skl_graph_t, /) -> tuple[Any, ...]:
    """"""
    slots = []
    for slot in graph.__slots__:
        slots.extend((slot, getattr(graph, slot)))

    # If it was not taken care of by json-any:
    # node_details = {_nde: _dtl for _nde, _dtl in graph.nodes.data("details")}
    # edge_details = {
    #     tuple(_dge): _dtl for *_dge, _dtl in graph.edges.data("details", keys=True)
    # }
    # Then add node_details and edge_details to the description.

    return tuple(slots) + (graph.AsNetworkX(),)


def NewFromJsonDescription(description: tuple[Any, ...], /) -> skl_graph_t:
    """"""
    output = skl_graph_t()

    *slots, graph = description

    for s_idx in range(0, slots.__len__(), 2):
        setattr(output, slots[s_idx], slots[s_idx + 1])

    output.add_edges_from(graph.edges(data=True, keys=True))
    attributes = {_nde: _ttr for _nde, _ttr in graph.nodes(data=True)}
    ntkx.set_node_attributes(output, attributes)

    # If it was not taken care of by json-any ( after retrieving node_details and
    # edge_details from the description).
    # for node, details in node_details.items():
    #     output.nodes[node]["details"] = details
    # for (node_0, node_1, uid), details in edge_details.items():
    #     output.edges[node_0, node_1, uid]["details"] = details

    return output


if __name__ == "__main__":
    #
    import doctest

    doctest.testmod()



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Sequence

import numpy as nmpy

from skl_graph.type.edge import edge_t
from skl_graph.type.node import branch_node_t, end_node_t


def Issues(
    edge: edge_t,
    source: end_node_t | branch_node_t,
    target: end_node_t | branch_node_t,
    uid: str,
    /,
) -> Sequence[str]:
    """"""
    # TODO: Check all attributes values, like with nodes, and do it first.
    output = _NodeIndependentIssues(edge)

    if not _HasValidEndSites(edge, source, target):
        output.append(f"{uid}: Invalid end sites")

    return output


def _NodeIndependentIssues(edge: edge_t, /) -> list[str]:
    """Partial validity test.

    Missing validity test: The end sites must be found in the adjacent nodes position/sites, which can only be
    tested at the graph level. See _HasValidEndSites.
    """
    output = []

    n_sites = edge.n_sites
    length = edge.lengths.length
    segment_sq_lengths = edge.lengths.segment_sq_lengths
    if length < n_sites - 1:
        output.append(
            f"{edge.uid}: {length}: Computed length cannot be smaller "
            f"than number of sites - 1={n_sites - 1}"
        )
    if nmpy.any(segment_sq_lengths == 0):
        output.append(f"{edge.uid}: Repeated sites")
    if nmpy.any(segment_sq_lengths > edge.dim):
        output.append(f"{edge.uid}: Site gaps")

    return output


def _HasValidEndSites(
    edge: edge_t,
    source: end_node_t | branch_node_t,
    target: end_node_t | branch_node_t,
    /,
) -> bool:
    """"""
    end_sites = tuple(
        tuple(edge.sites[_ddx][_sdx] for _ddx in range(edge.dim)) for _sdx in (0, -1)
    )
    n_found_end_sites = [0, 0]

    for node in (source, target):
        for e_idx in (0, 1):
            if isinstance(node, end_node_t):
                end_site_found = nmpy.array_equal(end_sites[e_idx], node.position)
            else:
                end_site = nmpy.reshape(end_sites[e_idx], (edge.dim, 1))
                end_site_found = any(
                    nmpy.all(end_site == nmpy.array(node.sites), axis=0)
                )
            if end_site_found:
                n_found_end_sites[e_idx] += 1

    if source.uid == target.uid:
        n_expected = 2
    else:
        n_expected = 1

    return n_found_end_sites == [n_expected, n_expected]



# Copyright CNRS/Inria/UNS
# Contributor(s): Eric Debreuve (since 2018)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import dataclasses as dtcl
from enum import Enum as enum_t
from typing import ClassVar

node_style_raw_h = str  # color
node_styles_raw_h = dict[int | str | None, node_style_raw_h]
node_either_raw_h = node_style_raw_h | node_styles_raw_h

edge_style_raw_h = tuple[str, int, str]  # type, size, color
edge_styles_raw_h = tuple[edge_style_raw_h, edge_style_raw_h]

# show, size, color | show (False)
label_style_raw_h = tuple[bool, float, str] | bool
label_styles_raw_h = tuple[label_style_raw_h, label_style_raw_h]

# show, size, color | show (False)
direction_style_raw_h = tuple[bool, int, str] | bool


class plot_mode_e(enum_t):
    Networkx = 0
    SKL_Pixel = 1  # Exact, pixel-wise edges
    SKL_Curve = 2  # Pchip (Piecewise Cubic Hermite Interpolating Polynomial) approximation of edges


@dtcl.dataclass(slots=True, repr=False, eq=False)
class node_style_t:
    color: str  # E.g., "b"

    DEFAULT_COLORS: ClassVar[tuple[tuple[int | None, str]]] = ((None, "g"), (1, "r"))

    @classmethod
    def Default(cls) -> node_styles_h:
        """"""
        return {_key: cls(color=_vle) for _key, _vle in cls.DEFAULT_COLORS}

    @classmethod
    def NewFromUnstructured(cls, style: node_either_raw_h, /) -> node_styles_h:
        """"""
        if isinstance(style, dict):  # node_styles_raw_h
            # _dgr=degree or a string that must be present in the node property keys.
            output = {_dgr: cls(color=_stl) for _dgr, _stl in style.items()}
        else:
            output = {None: cls(color=style)}

        return output


@dtcl.dataclass(slots=True, repr=False, eq=False)
class edge_style_t:
    type: str  # E.g., "-"
    size: int  # "linewidth"
    color: str  # E.g., "b"

    DEFAULT_TYPES: ClassVar[tuple[str, str]] = ("-", ":")  # Regular, self-loop
    DEFAULT_SIZE: ClassVar[int] = 2
    DEFAULT_COLOR: ClassVar[str] = "k"

    @classmethod
    def Default(cls) -> edge_styles_h:
        """"""
        # noinspection PyTypeChecker
        return tuple(
            cls(type=_typ, size=cls.DEFAULT_SIZE, color=cls.DEFAULT_COLOR)
            for _typ in cls.DEFAULT_TYPES
        )

    @classmethod
    def AllFromUnstructured(
        cls,
        current: edge_styles_h,
        /,
        *,
        edge: edge_style_raw_h = None,
        regular_edge: edge_style_raw_h = None,
        self_loop: edge_style_raw_h = None,
        edges: edge_styles_raw_h = None,
    ) -> edge_styles_h | None:
        """"""
        return _AllFromUnstructured(
            edge_style_t,
            current,
            common=edge,
            first=regular_edge,
            last=self_loop,
            both=edges,
        )

    @classmethod
    def NewFromUnstructured(cls, style: edge_style_raw_h, /) -> edge_style_t:
        """"""
        return cls(type=style[0], size=style[1], color=style[2])


@dtcl.dataclass(slots=True, repr=False, eq=False)
class label_style_t:
    show: bool
    size: float  # "fontsize"
    color: str  # "k"

    DEFAULT_SIZE: ClassVar[float] = 6.0
    DEFAULT_COLOR: ClassVar[str] = "k"

    @classmethod
    def Default(cls) -> label_styles_h:
        """"""
        return (
            cls.DefaultForNodes(),
            cls.DefaultForEdges(),
        )

    @classmethod
    def DefaultForNodes(cls) -> label_style_t:
        """"""
        return cls.NewFromUnstructured(True)

    @classmethod
    def DefaultForEdges(cls) -> label_style_t:
        """"""
        return cls.NewFromUnstructured(False)

    @classmethod
    def AllFromUnstructured(
        cls,
        current: label_styles_h,
        /,
        *,
        label: label_style_raw_h = None,
        node: label_style_raw_h = None,
        edge: label_style_raw_h = None,
        labels: label_styles_raw_h = None,
    ) -> label_styles_h | None:
        """"""
        return _AllFromUnstructured(
            label_style_t,
            current,
            common=label,
            first=node,
            last=edge,
            both=labels,
        )

    @classmethod
    def NewFromUnstructured(cls, style: label_style_raw_h, /) -> label_style_t:
        """"""
        if isinstance(style, bool):
            show = style
            size = cls.DEFAULT_SIZE
            color = cls.DEFAULT_COLOR
        else:
            show, size, color = style

        return cls(show=show, size=size, color=color)


@dtcl.dataclass(slots=True, repr=False, eq=False)
class direction_style_t:
    show: bool
    size: int  # "linewidth"
    color: str  # "k"

    DEFAULT_SIZE: ClassVar[int] = 2
    DEFAULT_COLOR: ClassVar[str] = "y"

    @classmethod
    def Default(cls) -> direction_style_t:
        """"""
        return cls.NewFromUnstructured(False)

    @classmethod
    def NewFromUnstructured(cls, style: direction_style_raw_h, /) -> direction_style_t:
        """"""
        if isinstance(style, bool):
            show = style
            size = cls.DEFAULT_SIZE
            color = cls.DEFAULT_COLOR
        else:
            show, size, color = style

        return cls(show=show, size=size, color=color)


node_styles_h = dict[int | str | None, node_style_t]
edge_styles_h = tuple[edge_style_t, edge_style_t]  # regular edges, self-loops
label_styles_h = tuple[label_style_t, label_style_t]  # nodes, edges


def _AllFromUnstructured(
    class_: edge_style_t.__class__ | label_style_t.__class__,
    current: edge_styles_h | label_styles_h,
    /,
    *,
    common: edge_style_raw_h | label_styles_raw_h = None,
    first: edge_style_raw_h | label_styles_raw_h = None,
    last: edge_style_raw_h | label_styles_raw_h = None,
    both: edge_styles_raw_h | label_styles_raw_h = None,
) -> edge_styles_h | label_styles_h | None:
    """"""
    if common is not None:
        style = class_.NewFromUnstructured(common)
        output = (style, style)
    elif first is not None:
        style = class_.NewFromUnstructured(first)
        output = (style, current[1])
    elif last is not None:
        style = class_.NewFromUnstructured(last)
        output = (current[0], style)
    elif both is not None:
        output = tuple(map(class_.NewFromUnstructured, both))
    else:
        output = None

    return output
